import ast
import json
import uuid


from ivoryos.utils.global_config import GlobalConfig

global_config = GlobalConfig()

if global_config.building_blocks:
    building_blocks = {
        inner_key: f"{block_key}.{inner_key}"
        for block_key, block_value in global_config.building_blocks.items()
        for inner_key in block_value.keys()
    }
else:
    building_blocks = {}

def generate_uuid():
    return int(str(uuid.uuid4().int)[:15])

def infer_type(value):
    if isinstance(value, bool):
        return "bool"
    elif isinstance(value, int):
        return "int"
    elif isinstance(value, float):
        return "float"
    elif isinstance(value, str):
        return "str"
    elif isinstance(value, (ast.Name, str)) and str(value).startswith("#"):
        return "float"  # default fallback for variables
    else:
        return "unknown"

def convert_to_cards(source_code: str):
    tree = ast.parse(source_code)
    cards = []
    card_id = 1
    block_stack = []  # to track control flow UUIDs

    def new_id():
        nonlocal card_id
        val = card_id
        card_id += 1
        return val

    def add_card(card):
        cards.append(card)

    def is_supported_assignment(node):
        return (
            isinstance(node.targets[0], ast.Name) and
            isinstance(node.value, ast.Constant)
        )

    class CardVisitor(ast.NodeVisitor):
        def __init__(self):
            self.defined_types = {}  # <-- always exists


        def visit_FunctionDef(self, node):
            self.defined_types = {
                arg.arg: ast.unparse(arg.annotation) if arg.annotation else "float"
                for arg in node.args.args
            }
            for stmt in node.body:
                self.visit(stmt)

        def visit_If(self, node):
            uuid_ = generate_uuid()
            block_stack.append(("if", uuid_))

            add_card({
                "action": "if",
                "arg_types": {"statement": ""},
                "args": {"statement": ast.unparse(node.test)},
                "id": new_id(),
                "instrument": "if",
                "return": "",
                "uuid": uuid_
            })

            for stmt in node.body:
                self.visit(stmt)

            if node.orelse:
                add_card({
                    "action": "else",
                    "args": {},
                    "id": new_id(),
                    "instrument": "if",
                    "return": "",
                    "uuid": uuid_
                })
                for stmt in node.orelse:
                    self.visit(stmt)

            _, block_uuid = block_stack.pop()
            add_card({
                "action": "endif",
                "args": {},
                "id": new_id(),
                "instrument": "if",
                "return": "",
                "uuid": block_uuid
            })

        def visit_While(self, node):
            uuid_ = generate_uuid()
            block_stack.append(("while", uuid_))

            add_card({
                "action": "while",
                "arg_types": {"statement": ""},
                "args": {"statement": ast.unparse(node.test)},
                "id": new_id(),
                "instrument": "while",
                "return": "",
                "uuid": uuid_
            })

            for stmt in node.body:
                self.visit(stmt)

            _, block_uuid = block_stack.pop()
            add_card({
                "action": "endwhile",
                "args": {},
                "id": new_id(),
                "instrument": "while",
                "return": "",
                "uuid": block_uuid
            })

        def visit_Assign(self, node):
            if is_supported_assignment(node):
                var_name = node.targets[0].id
                value = node.value.value
                add_card({
                    "action": var_name,
                    "arg_types": {"statement": infer_type(value)},
                    "args": {"statement": value},
                    "id": new_id(),
                    "instrument": "variable",
                    "return": "",
                    "uuid": generate_uuid()
                })
            elif isinstance(node.value, ast.Await):
                self.handle_call(node.value.value, ret_var=node.targets[0].id, awaited=True)

            elif isinstance(node.value, ast.Call):
                self.handle_call(node.value, ret_var=node.targets[0].id)

        def visit_Expr(self, node):
            if isinstance(node.value, ast.Await):
                # node.value is ast.Await
                self.handle_call(node.value.value, awaited=True)
            elif isinstance(node.value, ast.Call):
                self.handle_call(node.value)

        def handle_call(self, node, ret_var="", awaited=False):
            func_parts = []
            f = node.func
            while isinstance(f, ast.Attribute):
                func_parts.insert(0, f.attr)
                f = f.value
            if isinstance(f, ast.Name):
                func_parts.insert(0, f.id)

            full_func_name = ".".join(func_parts)

            # Check if it starts with 'deck.' and strip it
            if full_func_name.startswith("deck."):
                remaining_parts = full_func_name.split(".")[1:]
                # Must have at least instrument.action
                if len(remaining_parts) >= 2:
                    instrument = ".".join(remaining_parts[:-1])
                    action = remaining_parts[-1]
                else:
                    return # invalid format like deck.something
            elif full_func_name.startswith("blocks."):
                instrument = ".".join(func_parts[:-1])
                action = func_parts[-1]
            elif func_parts[-1] in building_blocks.keys():
                instrument = building_blocks.get(func_parts[-1])
                action = func_parts[-1]
            else:
                # Try to see if the first part is a known instrument in deck_snapshot
                # This covers cases like 'sdl.dose' where 'sdl' is the instrument
                if len(func_parts) >= 2:
                    possible_instrument = ".".join(func_parts[:-1])
                    possible_action = func_parts[-1]
                    if f"deck.{possible_instrument}" in global_config.deck_snapshot:
                        instrument = f"deck.{possible_instrument}"
                        action = possible_action
                    else:
                        return
                else:
                    return

            # # Validate that the instrument (if not a block or time) exists in deck_snapshot
            # if not instrument.startswith("blocks.") and instrument != "time":
            #      if instrument not in global_config.deck_snapshot:
            #          return # Not a valid instrument call for this deck
            #


            # --- special case for time.sleep ---
            if instrument == "time" and action == "sleep":
                wait_value = None
                if node.args:
                    arg_node = node.args[0]
                    if isinstance(arg_node, ast.Constant):
                        wait_value = arg_node.value
                    elif isinstance(arg_node, ast.Name):
                        wait_value = f"#{arg_node.id}"
                    else:
                        wait_value = ast.unparse(arg_node)

                add_card({
                    "action": "wait",
                    "arg_types": {"statement": infer_type(wait_value)},
                    "args": {"statement": wait_value},
                    "id": new_id(),
                    "instrument": "wait",
                    "return": ret_var,
                    "uuid": generate_uuid()
                })
                return
            # -----------------------------------


            args = {}
            arg_types = {}

            # Resolve function signature
            sig_params = []
            func_def = None

            # 1. Check in deck_snapshot
            if instrument in global_config.deck_snapshot:
                func_info = global_config.deck_snapshot[instrument].get(action)
                if func_info and 'signature' in func_info:
                    sig_params = list(func_info['signature'].parameters.keys())
                    sig_params.remove("self")
                    print(sig_params)

            # 2. Check in building_blocks
            # instrument might be mapped name or actual block name
            # building_blocks keys are short names, values are mapped names (e.g. 'test' -> 'blocks.general.test')
            if not sig_params:
                # Check if instrument is a block name e.g. "blocks.general"
                if instrument.startswith("blocks."):
                    if instrument in global_config.building_blocks:
                        func_info = global_config.building_blocks[instrument].get(action)
                        if func_info and 'signature' in func_info:
                            sig_params = list(func_info['signature'].parameters.keys())

            # Handle positional arguments
            for i, arg_node in enumerate(node.args):
                if i < len(sig_params):
                    arg_name = sig_params[i]
                else:
                    # Fallback if we have more args than parameters or no signature found
                    arg_name = f"arg_{i}"

                if isinstance(arg_node, ast.Constant):
                    value = arg_node.value
                elif isinstance(arg_node, ast.Name):
                    value = f"#{arg_node.id}"
                else:
                    value = ast.unparse(arg_node)

                args[arg_name] = value
                arg_types[arg_name] = infer_type(value)


            for kw in node.keywords:
                if kw.arg is None and isinstance(kw.value, ast.Dict):
                    for k_node, v_node in zip(kw.value.keys, kw.value.values):
                        key = k_node.value if isinstance(k_node, ast.Constant) else ast.unparse(k_node)
                        if isinstance(v_node, ast.Constant):
                            value = v_node.value
                        elif isinstance(v_node, ast.Name):
                            value = f"#{v_node.id}"
                        else:
                            value = ast.unparse(v_node)
                        args[key] = value
                        arg_types[key] = infer_type(value)
                else:
                    if isinstance(kw.value, ast.Constant):
                        value = kw.value.value
                    elif isinstance(kw.value, ast.Name):
                        value = f"#{kw.value.id}"
                    else:
                        value = ast.unparse(kw.value)
                    args[kw.arg] = value
                    arg_types[kw.arg] = (
                        self.defined_types.get(kw.value.id, "float")
                        if isinstance(kw.value, ast.Name)
                        else infer_type(value)
                    )

            card = {
                "action": action,
                "arg_types": arg_types,
                "args": args,
                "id": new_id(),
                "instrument": instrument,
                "return": ret_var,
                "uuid": generate_uuid()
            }

            if awaited:
                card["coroutine"] = True  # mark as coroutine if awaited

            add_card(card)

    CardVisitor().visit(tree)
    return cards

def extract_functions_and_convert(source_code: str):
    """
    Parses source code, extracts each top-level function definition,
    and converts its body to cards.
    Returns a dict {func_name: cards_list}.
    """
    tree = ast.parse(source_code)
    results = {}
    
    for node in tree.body:
        if isinstance(node, ast.FunctionDef):
             # Extract the source code for this function
             # ast.get_source_segment is available in 3.8+
             func_source = ast.get_source_segment(source_code, node)
             if func_source:
                 cards = convert_to_cards(func_source)
                 results[node.name] = {"cards": cards, "source": func_source}
    return results


if __name__ == "__main__":
    test = '''def workflow_dynamic(solid_amount_mg, methanol_amount_ml):
    """
    SDL workflow: dose solid, add methanol, equilibrate, and analyze
    
    Args:
        solid_amount_mg (float): Amount of solid to dose in mg
        methanol_amount_ml (float): Amount of methanol to dose in ml
    
    Returns:
        dict: Results containing analysis data
    """
    # Step 1: Dose solid material
    deck.sdl.dose_solid(solid_amount_mg)
    
    # Step 2: Add methanol solvent
    deck.sdl.dose_solvent(solvent_name='Methanol', amount_in_ml=methanol_amount_ml)
    
    # Step 3: Equilibrate at room temperature (assuming ~23°C) for 20 seconds
    deck.sdl.equilibrate(temp=23.0, duration=20.0)
    
    # Step 4: Analyze the sample
    analysis_results = deck.sdl.analyze(param_1=1, param_2=2)
    
    # test block 
    result = blocks.general.test(**{'a': 1, 'b': 2})
    
    # Brief pause for system stability
    time.sleep(1.0)
    
    # Return only analysis results
    return {'analysis_results': analysis_results}
    '''
    from pprint import pprint
    pprint(json.dumps(convert_to_cards(test)))


import asyncio
import inspect
from unittest.mock import MagicMock, patch

import pytest
from flask import Flask
from wtforms import StringField, FloatField

from ivoryos.utils import utils
from ivoryos.utils.db_models import Script
from ivoryos.utils.form import create_form_from_pseudo
from ivoryos.utils.task_runner import TaskRunner
from ivoryos.utils.utils import _inspect_class


# --- Fixtures & Helpers ---

@pytest.fixture
def app():
    app = Flask(__name__)
    app.config['WTF_CSRF_ENABLED'] = False
    return app

@pytest.fixture
def mock_app():
    app = MagicMock()
    app.app_context.return_value.__enter__.return_value = app
    return app

@pytest.fixture
def runner_with_mock_db():
    with patch('ivoryos.utils.task_runner.db') as mock_db:
        runner = TaskRunner()
        yield runner

class MockInstrument:
    def __init__(self):
        self._temp = 25.0

    @property
    def temperature(self):
        return self._temp

    @temperature.setter
    def temperature(self, value):
        self._temp = float(value)

class MockDeck:
    def __init__(self):
        self.inst = MockInstrument()

class TestDevice:
    def __init__(self):
        self._temp = 25.0

    def connect(self):
        """Connect to device"""
        pass

    @property
    def temperature(self):
        """Get temperature"""
        return self._temp

    @temperature.setter
    def temperature(self, value):
        self._temp = value

    @property
    def ready(self):
        """Is ready"""
        return True

# --- Test Classes ---

class TestPropertyIntrospection:
    """Tests from test_utils_properties.py"""
    
    def test_inspect_class_properties(self):
        device = TestDevice()
        result = _inspect_class(device)
        
        # Check regular method
        assert 'connect' in result
        assert result['connect']['docstring'] == 'Connect to device'
        assert not result['connect'].get('is_property')

        # Check property with setter
        assert 'temperature' in result
        temp_info = result['temperature']
        assert temp_info['is_property'] is True
        assert temp_info['has_setter'] is True
        assert temp_info['docstring'] == 'Get temperature'

        # Check read-only property
        assert 'ready' in result
        ready_info = result['ready']
        assert ready_info['is_property'] is True
        assert ready_info['has_setter'] is False
        assert ready_info['docstring'] == 'Is ready'

class TestPropertyCompilation:
    """Tests from test_compilation_properties.py"""

    def test_compile_property_setter(self):
        # Mock snapshot
        snapshot = {
            'my_inst': {
                'temperature': {
                    'is_property': True,
                    'has_setter': True
                },
                'connect': {
                    'is_property': False
                }
            }
        }

        # Create script with setter action
        script = Script(name="test_script")
        script.script_dict = {
            'prep': [],
            'script': [
                {
                    "id": 1,
                    "instrument": "my_inst",
                    "action": "temperature_(setter)",
                    "args": {"value": 30.5},
                    "return": "", # Setters have empty return
                    "arg_types": {'value': 'float'}
                },
                 {
                    "id": 2,
                    "instrument": "my_inst",
                    "action": "connect",
                    "args": {},
                    "return": "",
                },
                {
                    "id": 3,
                    "instrument": "my_inst",
                    "action": "temperature",
                    "args": {},
                    "return": "temp_val",
                    "arg_types": {},
                }
            ],
            'cleanup': []
        }
        
        # Compile
        # We expect 'my_inst.temperature = 30.5' and 'my_inst.connect()'
        # And 'temp_val = my_inst.temperature' (getter, no parens)
        
        exec_str_collection = script.compile(snapshot=snapshot)
        script_code = exec_str_collection['script']
        
        print(script_code)
        
        assert "my_inst.temperature = 30.5" in script_code
        assert "my_inst.connect()" in script_code
        assert "temp_val = my_inst.temperature" in script_code
        assert "my_inst.temperature()" not in script_code
        assert "my_inst.set_temperature" not in script_code

    def test_compile_property_setter_variable(self):
        # Mock snapshot
        snapshot = {
            'my_inst': {
                'temperature': {'is_property': True, 'has_setter': True}
            }
        }
        script = Script(name="test_script_var")
        script.script_dict = {
            'prep': [],
            'script': [
                 {
                    "id": 1,
                    "instrument": "variable",
                    "action": "my_var",
                    "args": {"statement": 10},
                     "return": "",
                     "arg_types": {"statement": "int"}
                },
                {
                    "id": 2,
                    "instrument": "my_inst",
                    "action": "temperature_(setter)",
                    "args": {"value": "#my_var"},
                    "return": "", 
                    "arg_types": {'value': 'float'}
                }
            ],
            'cleanup': []
        }
        
        exec_str_collection = script.compile(snapshot=snapshot)
        script_code = exec_str_collection['script']
        print(script_code)
        
        assert "my_inst.temperature = my_var" in script_code

class TestDesignPropertyHandler:
    """Tests from test_design_property_handler.py"""

    def test_get_arg_type_with_virtual_setter(self):
        # Simulate the logic I added to methods_handler
        
        # Mock definitions
        functions = {
            'temperature': {
                'is_property': True, 
                'signature': inspect.Signature(return_annotation=float)
            }
        }
        function_name = 'temperature_(setter)'
        kwargs = {'value': '30.5', 'hidden_name': 'temperature_(setter)'}
        
        # Simulate the look up logic
        function_data = functions.get(function_name)
        if not function_data and function_name.endswith("_(setter)"):
            prop_name = function_name[:-9]
            prop_data = functions.get(prop_name)
            if prop_data and prop_data.get('is_property'):
                sig = prop_data.get('signature')
                param_type = inspect._empty
                if sig and sig.return_annotation is not inspect._empty:
                    param_type = sig.return_annotation
                
                setter_sig = inspect.Signature(
                    parameters=[inspect.Parameter('value', inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=param_type)],
                    return_annotation=None
                )
                function_data = {'signature': setter_sig}
                
        assert function_data is not None
        assert 'signature' in function_data
        sig = function_data['signature']
        assert 'value' in sig.parameters
        assert sig.parameters['value'].annotation == float
        
        # Verify get_arg_type works with this synthesized data
        arg_types = utils.get_arg_type(kwargs, function_data)
        # create_form_from_pseudo uses 'value' as parameter name
        assert arg_types['value'] == 'float'

class TestPropertyForms:
    """Tests from test_form_properties.py"""

    def test_form_from_properties(self, app):
        with app.app_context():
            # Mock pseudo deck data
            pseudo = {
                'connect': {
                    'signature': inspect.Signature(),
                    'docstring': 'Connect',
                    'is_property': False
                },
                'temperature': {
                    'signature': inspect.Signature(return_annotation=float),
                    'docstring': 'Get temp',
                    'is_property': True,
                    'has_setter': True
                },
                'ready': {
                    'signature': inspect.Signature(return_annotation=bool),
                    'docstring': 'Is ready',
                    'is_property': True,
                    'has_setter': False
                }
            }
            
            forms = create_form_from_pseudo(pseudo, autofill=False, design=True)
            
            assert 'connect' in forms
            assert 'temperature' in forms
            assert 'temperature_(setter)' in forms
            assert 'ready' in forms
            assert 'ready_(setter)' not in forms

            # Check setter form fields
            setter_form = forms['temperature_(setter)']
            # Should have a 'value' field (derived from synthetic signature)
            assert hasattr(setter_form, 'value')
            # Should NOT have a 'return' field
            assert not hasattr(setter_form, 'return')
            # Since we passed design=True, fields are VariableOr...Field, but base depends on annotation
            # checks logic in create_form_for_method

class TestPropertyExecution:
    """Tests from test_task_runner_properties.py"""

    @pytest.mark.asyncio
    async def test_task_runner_properties(self, mock_app, runner_with_mock_db):
        deck = MockDeck()
        
        # Setup global config
        from ivoryos.utils.global_config import GlobalConfig
        gc = GlobalConfig()
        gc.defined_variables['my_inst'] = deck.inst
        gc.deck = deck
        
        # Test Getter
        res = await runner_with_mock_db.run_single_step("my_inst", "temperature", {}, current_app=mock_app)
        assert res['success'] is True, f"Getter failed: {res.get('output')}"
        assert res['output'] == 25.0

        # Test Setter
        res_set = await runner_with_mock_db.run_single_step("my_inst", "temperature_(setter)", {"value": 30.0}, current_app=mock_app)
        assert res_set['success'] is True, f"Setter failed: {res_set.get('output')}"
        assert deck.inst.temperature == 30.0



from unittest.mock import patch

from tests.conftest import TestEnum


def test_int_conversion(auth, test_deck):
    """Tests that a string from a form is converted to an integer."""
    with patch('ivoryos.control.routes.global_config.deck_instance.deck_dummy.int_method') as mock_method:
        auth.post('/ivoryos/control/deck.dummy/call/int_method', data={'arg': '123'})
        # Check that the mock was called with an integer
        mock_method.assert_called_with(arg=123)

def test_float_conversion(auth, test_deck):
    """Tests that a string from a form is converted to a float."""
    with patch('ivoryos.control.routes.global_config.deck_instance.deck_dummy.float_method') as mock_method:
        auth.post('/ivoryos/control/deck.dummy/call/float_method', data={'arg': '123.45'})
        # Check that the mock was called with a float
        mock_method.assert_called_with(arg=123.45)

def test_bool_conversion(auth, test_deck):
    """Tests that a string from a form is converted to a boolean."""
    with patch('ivoryos.control.routes.global_config.deck_instance.deck_dummy.bool_method') as mock_method:
        # Test with 'true'
        auth.post('/ivoryos/control/deck.dummy/call/bool_method', data={'arg': 'true'})
        mock_method.assert_called_with(arg=True)
        # Test with 'false'
        auth.post('/ivoryos/control/deck.dummy/call/bool_method', data={'arg': 'false'})
        mock_method.assert_called_with(arg=False)

def test_list_conversion(auth, test_deck):
    """Tests that a comma-separated string from a form is converted to a list."""
    with patch('ivoryos.control.routes.global_config.deck_instance.deck_dummy.list_method') as mock_method:
        auth.post('/ivoryos/control/deck.dummy/call/list_method', data={'arg': 'a,b,c'})
        # Check that the mock was called with a list of strings
        mock_method.assert_called_with(arg=['a', 'b', 'c'])

def test_enum_conversion(auth, test_deck):
    """Tests that a string from a form is converted to an Enum member."""
    with patch('ivoryos.control.routes.global_config.deck_instance.deck_dummy.enum_method') as mock_method:
        auth.post('/ivoryos/control/deck.dummy/call/enum_method', data={'arg': 'OPTION_B'})
        # Check that the mock was called with the correct Enum member
        mock_method.assert_called_with(arg=TestEnum.OPTION_B)


from flask import Blueprint, redirect, url_for, flash, request, render_template, session
from flask_login import login_required, login_user, logout_user, LoginManager, current_user
import bcrypt
from sqlalchemy_utils.types import password

from ivoryos.utils.db_models import Script, User, db
from ivoryos.utils.utils import post_script_file
login_manager = LoginManager()
# from flask import g

auth = Blueprint('auth', __name__, template_folder='templates')


@auth.route('/login', methods=['GET', 'POST'])
def login():
    """
    .. :quickref: User; login user

    .. http:get:: /auth/login

    load user login form.

    .. http:post:: /auth/login

    :form username: username
    :form password: password
    :status 302: and then redirects to homepage
    :status 401: incorrect password, redirects to :http:get:`/ivoryos/auth/login`
    """
    if request.method == 'POST':
        username = request.form.get('username')
        password = request.form.get('password')

        # session.query(User, User.name).all()
        user = db.session.query(User).filter(User.username == username).first()
        input_password = password.encode('utf-8')
        # if user and bcrypt.checkpw(input_password, user.hashPassword.encode('utf-8')):
        if user and bcrypt.checkpw(input_password, user.hashPassword):
            # password.encode("utf-8")
            # user = User(username, password.encode("utf-8"))
            login_user(user)
            # g.user = user
            # session['user'] = username
            script_file = Script(author=username)
            session["script"] = script_file.as_dict()
            session['hidden_functions'], session['card_order'], session['prompt'] = {}, {}, {}
            session['autofill'] = False
            post_script_file(script_file)
            return redirect(url_for('main.index'))
        else:
            flash("Incorrect username or password")
            return redirect(url_for('auth.login'))
    return render_template('login.html')


@auth.route('/signup', methods=['GET', 'POST'])
def signup():
    """
    .. :quickref: User; signup for a new account

    .. http:get:: /auth/signup

    load user sighup

    .. http:post:: /auth/signup

    :form username: username
    :form password: password
    :status 302: and then redirects to :http:get:`/ivoryos/auth/login`
    :status 409: when user already exists, redirects to :http:get:`/ivoryos/auth/signup`
    """
    if request.method == 'POST':
        username = request.form.get('username')
        password = request.form.get('password')

        # Query the database to see if the user already exists.
        existing_user = User.query.filter_by(username=username).first()

        if existing_user:
            flash("User already exists :(", "error")
            return render_template('signup.html'), 409
        hashed = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt())
        user = User(username, hashed)
        db.session.add(user)
        db.session.commit()
        return redirect(url_for('auth.login'))
    return render_template('signup.html')

@auth.route("/change-password", methods=['GET', 'POST'])
@login_required
def change_password():
    """
    .. :quickref: User; change password

    .. http:get:: /auth/change-password

    .. http:post:: /auth/change-password

    change password
    """
    if request.method == "POST":
        old_password = request.form.get("old_password")
        new_password = request.form.get("new_password")
        # confirm_password = request.form.get("confirm_password")
        user = User.query.filter_by(username=current_user.get_id()).first()
        if not bcrypt.checkpw(old_password.encode('utf-8'), user.hashPassword):
            flash("Incorrect password")
            return redirect(url_for("auth.change_password"))
        user.hashPassword = bcrypt.hashpw(new_password.encode('utf-8'), bcrypt.gensalt())
        db.session.commit()
        return redirect(url_for("main.index"))
    return render_template("change_password.html")

@auth.route("/logout")
@login_required
def logout():
    """
    .. :quickref: User; logout the user

    .. http:get:: /auth/logout

    logout the current user, clear session info, and redirect to the login page.
    """
    logout_user()
    session.clear()
    return redirect(url_for('auth.login'))


def test_design_page_loads_for_auth_user(auth):
    """
    GIVEN an authenticated user
    WHEN the design page is accessed
    THEN the page should load successfully
    """
    response = auth.get('/ivoryos/design/script/', follow_redirects=True)
    assert response.status_code == 200
    assert b'<title>IvoryOS | Design</title>' in response.data # Assuming this text exists


def test_clear_canvas(auth):
    """
    Tests clearing the design canvas.
    """
    response = auth.get('/ivoryos/design/clear', follow_redirects=True)
    assert response.status_code == 200
    # assert b'Operations' in response.data

# def test_add_action(auth, test_deck):
#     """
#     Tests adding an action to the design canvas.
#     """
#     response = auth.post('/ivoryos/design/script/deck.dummy/', data={
#         'hidden_name': 'int_method',
#         'arg': '10'
#     }, follow_redirects=True)
#     assert response.status_code == 200

def test_experiment_run_page(auth):
    """
    Tests the experiment run page.
    """
    response = auth.get('/ivoryos/design/campaign')
    assert response.status_code == 200
    assert b'Run Panel' in response.data


from enum import Enum

import bcrypt
import pytest
from ivoryos.config import get_config

from ivoryos import create_app, socketio, db as _db, utils, global_config
from ivoryos.utils.db_models import User


@pytest.fixture(scope='session')
def app():
    """Create a new app instance for the test session."""
    _app = create_app(get_config('testing'))
    return _app

@pytest.fixture
def client(app):
    """A test client for the app."""
    with app.test_client() as client:
        with app.app_context():
            _db.create_all()
        yield client
        with app.app_context():
            _db.drop_all()

# @pytest.fixture(scope='session')
# def db(app):
#     """Session-wide test database."""
#     with app.app_context():
#         _db.create_all()
#         yield _db
#         _db.drop_all()

@pytest.fixture(scope='module')
def init_database(app):
    """
    Creates the database tables and seeds it with a default test user.
    This runs once per test module.
    """
    with app.app_context():
        # Drop everything first to ensure a clean slate
        _db.drop_all()
        # Create the database tables
        _db.create_all()

        # Insert a default user for authentication tests
        # Note: In a real app with password hashing, you'd call a hash function here.
        password = 'password'
        bcrypt_password = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt())
        default_user = User(username='testuser', password=bcrypt_password)
        _db.session.add(default_user)
        _db.session.commit()

        yield _db  # this is where the testing happens!

        # Teardown: drop all tables after the tests in the module are done
        _db.drop_all()


# ---------------------
# Authentication Fixture
# ---------------------

@pytest.fixture(scope='function')
def auth(client, init_database):
    """
    Logs in the default user for a single test function.
    Depends on `init_database` to ensure the user exists.
    Handles logout as part of teardown.
    """
    # Log in the default user
    client.post('/ivoryos/auth/login', data={
        'username': 'testuser',
        'password': 'password'
    }, follow_redirects=True)

    yield client  # this is where the testing happens!

    # Log out the user after the test is done
    client.get('/ivoryos/auth/logout', follow_redirects=True)


@pytest.fixture
def socketio_client(app):
    """A test client for Socket.IO."""
    return socketio.test_client(app)


class TestEnum(Enum):
    """An example Enum for testing type conversion."""
    OPTION_A = 'A'
    OPTION_B = 'B'

class DummyModule:
    """A more comprehensive dummy instrument for testing."""
    def int_method(self, arg: int = 1):
        return arg

    def float_method(self, arg: float = 1.0):
        return arg

    def bool_method(self, arg: bool = False):
        return arg

    def list_method(self, arg: list = None):
        return arg or []

    def enum_method(self, arg: TestEnum = TestEnum.OPTION_A):
        return arg

    def str_method(self) -> dict:
        return {'status': 'OK'}


@pytest.fixture
def test_deck(app):
    """
    A fixture that creates and loads a predictable 'deck' of dummy instruments
    for testing the dynamic control routes.
    """
    dummy_module = DummyModule()
    snapshot = utils.create_deck_snapshot(dummy_module)

    with app.app_context():
        global_config.deck_snapshot = snapshot
        global_config.deck = dummy_module # instantiate the class

    yield DummyModule

    with app.app_context():
        global_config.deck_snapshot = {}
        global_config.deck = {}


from enum import Enum, EnumMeta
from typing import Union, Any
try:
    from typing import get_origin, get_args
except ImportError:
    # For Python versions = 3.7, use typing_extensions
    from typing_extensions import get_origin, get_args

from wtforms.fields.choices import SelectField
from wtforms.fields.core import Field
from wtforms.validators import InputRequired, ValidationError, Optional
from wtforms.widgets.core import TextInput

from flask_wtf import FlaskForm
from wtforms import StringField, FloatField, HiddenField, BooleanField, IntegerField
import inspect

from ivoryos.utils.db_models import Script
from ivoryos.utils.global_config import GlobalConfig

global_config = GlobalConfig()

def find_variable(data, script):
    """
    find user defined variables and return values in the script:Script
    :param data: string of input variable name
    :param script:Script object
    """
    variables: dict[str, str] = script.get_variables()
    for variable_name, variable_type in variables.items():
        if variable_name == data:
            return data, variable_type  # variable_type int float str or "function_output"
    return None, None


class VariableOrStringField(Field):
    widget = TextInput()

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrStringField, self).__init__(label, validators, **kwargs)
        self.script = script

    def process_formdata(self, valuelist):
        if valuelist:
            if not self.script.editing_type == "script" and valuelist[0].startswith("#"):
                raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
            self.data = valuelist[0]

    def _value(self):
        if self.script:
            variable, variable_type = find_variable(self.data, self.script)
            if variable:
                return variable
        # print("_value", self.data, type(self.data))
        return str(self.data) if self.data else ''


class VariableOrFloatField(Field):
    widget = TextInput()

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrFloatField, self).__init__(label, validators, **kwargs)
        self.script = script

    def _value(self):
        if self.script:
            variable, variable_type = find_variable(self.data, self.script)
            if variable:
                return variable

        if self.raw_data:
            return self.raw_data[0]
        if self.data is not None:
            return str(self.data)
        return ""

    def process_formdata(self, valuelist):
        if not valuelist:
            return
        elif valuelist[0].startswith("#"):
            if not self.script.editing_type == "script":
                raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
            self.data = valuelist[0]
            return
        try:
            if self.script:
                try:
                    variable, variable_type = find_variable(valuelist[0], self.script)
                    if variable:
                        if not variable_type == "function_output":
                            if variable_type not in ["float", "int"]:
                                raise ValueError("Variable is not a valid float")
                        self.data = variable
                        return
                except ValueError:
                    pass
            self.data = float(valuelist[0])
        except ValueError as exc:
            self.data = None
            raise ValueError(self.gettext("Not a valid float value.")) from exc


# unset_value = UnsetValue()


class VariableOrIntField(Field):
    widget = TextInput()

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrIntField, self).__init__(label, validators, **kwargs)
        self.script = script

    def _value(self):
        if self.script:
            variable, variable_type = find_variable(self.data, self.script)
            if variable:
                return variable

        if self.raw_data:
            return self.raw_data[0]
        if self.data is not None:
            return str(self.data)
        return ""

    def process_formdata(self, valuelist):
        if not valuelist:
            return
        if self.script:
            variable, variable_type = find_variable(valuelist[0], self.script)
            if variable:
                try:
                    if not variable_type == "function_output":
                        if not variable_type == "int":
                            raise ValueError("Not a valid integer value")
                    self.data = str(variable)
                    return
                except ValueError:
                    pass
        if valuelist[0].startswith("#"):
            if not self.script.editing_type == "script":
                raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
            self.data = valuelist[0]
            return
        if valuelist[0] == "":
            # print("empty input", valuelist)
            self.data = None
            return
        try:
            self.data = int(valuelist[0])
        except ValueError as exc:
            self.data = None
            raise ValueError(self.gettext("Not a valid integer value.")) from exc


class VariableOrBoolField(Field):
    widget = TextInput()
    false_values = (False, "false", "", "False", "f", "F", "n")

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrBoolField, self).__init__(label, validators, **kwargs)
        self.script = script

    def process_data(self, value):

        if self.script:
            variable, variable_type = find_variable(value, self.script)
            if variable:
                if not variable_type == "function_output":
                    raise ValueError("Not accepting boolean variables")
                return variable
        if isinstance(value, str) and value.startswith("#"):
            self.data = value
            return value
        if value in self.false_values:
            self.data = False
        else:
            self.data = True
        return None

    def process_formdata(self, valuelist):
        # todo
        # print(valuelist)
        if not valuelist or not type(valuelist) is list:
            self.data = False
        else:
            value = valuelist[0] if type(valuelist) is list else valuelist
            if value.startswith("#"):
                if not self.script.editing_type == "script":
                    raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
                self.data = valuelist[0]
            elif value in self.false_values:
                self.data = False
            else:
                self.data = True

    def _value(self):

        if self.script:
            variable, variable_type = find_variable(self.raw_data, self.script)
            if variable:
                return variable

        if self.raw_data:
            return str(self.raw_data[0])
        return str(self.data)


class FlexibleEnumField(StringField):
    def __init__(self, label=None, validators=None, choices=None, script=None, **kwargs):
        super().__init__(label, validators, **kwargs)
        self.script = script
        self.enum_class = self._resolve_enum(choices)
        self.choices = [e.name for e in self.enum_class]
        # self.value_list = [e.name for e in self.enum_class]

    def _resolve_enum(self, annotation):
        """Extract Enum from Enum or Optional[Enum]"""
        # Case: direct Enum
        if isinstance(annotation, EnumMeta):
            return annotation

        # Case: Optional / Union
        origin = get_origin(annotation)
        if origin is Union:
            for arg in get_args(annotation):
                if isinstance(arg, EnumMeta):
                    return arg

        raise TypeError(f"FlexibleEnumField expected Enum or Optional[Enum], got {annotation!r}")

    def process_formdata(self, valuelist):
        # todo right now enum types will be processed to methods as a str, so the method must convert to the enum type
        if valuelist:
            key = valuelist[0]
            # Treat empty string or "None" as null value
            if key in ("", None, "None"):
                self.data = None
                return

            if key in self.choices:
                # Convert the string key to Enum instance
                self.data = self.enum_class[key].value
            elif key.startswith("#"):
                if not self.script.editing_type == "script":
                    raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
                self.data = key
            else:
                raise ValidationError(
                    f"Invalid choice: '{key}'. Must match one of {list(self.enum_class.__members__.keys())}")



def parse_annotation(annotation):
    """
    Given a type annotation, return:
    - a list of all valid types (excluding NoneType)
    - a boolean indicating if the value can be None (optional)
    """
    origin = get_origin(annotation)
    args = get_args(annotation)

    if annotation is Any:
        return [str], True  # fallback: accept any string, optional

    if origin is Union:
        types = list(set(args))
        is_optional = type(None) in types
        non_none_types = [t for t in types if t is not type(None)]
        return non_none_types, is_optional

    # Not a Union, just a regular type
    return [annotation], False

def create_form_for_method(method, autofill, script=None, design=True):
    """
    Create forms for each method or signature
    :param method: dict(docstring, signature)
    :param autofill:bool if autofill is enabled
    :param script:Script object
    :param design: if design is enabled
    """

    class DynamicForm(FlaskForm):
        pass

    annotation_mapping = {
        int: (VariableOrIntField if design else IntegerField, 'Enter integer value'),
        float: (VariableOrFloatField if design else FloatField, 'Enter numeric value'),
        str: (VariableOrStringField if design else StringField, 'Enter text'),
        bool: (VariableOrBoolField if design else BooleanField, 'Empty for false')
    }
    sig = method if type(method) is inspect.Signature else inspect.signature(method)

    for param in sig.parameters.values():
        if param.name == 'self':
            continue
        # formatted_param_name = format_name(param.name)

        default_value = None
        if autofill:
            default_value = f'#{param.name}'
        else:
            if param.default is not param.empty:
                if isinstance(param.default, Enum):
                    default_value = param.default.name
                else:
                    default_value = param.default

        field_kwargs = {
            "label": param.name,
            "default": default_value,
            "validators": [InputRequired()] if param.default is param.empty else [Optional()],
            **({"script": script} if (autofill or design) else {})
        }
        if _is_enum_type(param.annotation):
            enum_class = _unwrap_enum_type(param.annotation)
            field_class = FlexibleEnumField
            placeholder_text = f"Choose or type a value for {enum_class.__name__} (start with # for custom)"

            extra_kwargs = {"choices": param.annotation}

        else:
            # print(param.annotation)
            annotation, optional = parse_annotation(param.annotation)
            annotation = annotation[0]
            field_class, placeholder_text = annotation_mapping.get(
                annotation,
                (VariableOrStringField if design else StringField, f'Enter {param.annotation} value')
            )
            extra_kwargs = {}
            if optional:
                field_kwargs["filters"] = [lambda x: x if x != '' else None]

            if annotation is bool:
                # Boolean fields should not use InputRequired
                field_kwargs["validators"] = []  # or [Optional()]
            else:
                field_kwargs["validators"] = [InputRequired()] if param.default is param.empty else [Optional()]

        render_kwargs = {"placeholder": placeholder_text}

        # Create the field with additional rendering kwargs for placeholder text
        field = field_class(**field_kwargs, render_kw=render_kwargs, **extra_kwargs)
        setattr(DynamicForm, param.name, field)

    # setattr(DynamicForm, f'add', fname)
    return DynamicForm


def _is_enum_type(tp):
    # Optional[Enum] comes through as Union[Enum, NoneType]
    origin = get_origin(tp)

    # Non-Optional direct enum
    if isinstance(tp, type) and issubclass(tp, Enum):
        return True

    # Optional/Union case → unwrap inner types
    if origin is Union:
        return any(
            isinstance(arg, type) and issubclass(arg, Enum)
            for arg in get_args(tp)
            if arg is not type(None)
        )

    return False

def _unwrap_enum_type(tp):
    from typing import get_origin, get_args, Union
    from enum import Enum

    # Bare Enum
    if isinstance(tp, type) and issubclass(tp, Enum):
        return tp

    # Optional/Union conversion
    origin = get_origin(tp)
    if origin is Union:
        for arg in get_args(tp):
            if arg is not type(None) and isinstance(arg, type) and issubclass(arg, Enum):
                return arg

    return None

def create_add_form(attr, attr_name, autofill: bool, script=None, design: bool = True):
    """
    Create forms for each method or signature
    :param attr: dict(docstring, signature)
    :param attr_name: method name
    :param autofill:bool if autofill is enabled
    :param script:Script object
    :param design: if design is enabled. Design allows string input for parameter names ("#param") for all fields
    """
    signature = attr.get('signature', {})
    docstring = attr.get('docstring', "")
    return_type = signature.return_annotation
    # print(signature, docstring)
    dynamic_form = create_form_for_method(signature, autofill, script, design)
    if design:
        if return_type is not None:
            return_value = StringField(label='Save value as', render_kw={"placeholder": "Optional"})
            setattr(dynamic_form, 'return', return_value)
        batch_action = BooleanField(label='Batch Action', render_kw={"placeholder": "Optional"})
        setattr(dynamic_form, 'batch_action', batch_action)
    hidden_method_name = HiddenField(name=f'hidden_name', description=docstring, render_kw={"value": f'{attr_name}'})
    setattr(dynamic_form, 'hidden_name', hidden_method_name)
    return dynamic_form


def create_form_from_module(sdl_module, autofill: bool = False, script=None, design: bool = False):
    """
    Create forms for each method, used for control routes
    :param sdl_module: method module
    :param autofill:bool if autofill is enabled
    :param script:Script object
    :param design: if design is enabled
    """
    method_forms = {}
    for attr_name in dir(sdl_module):
        try:
            method = getattr(sdl_module, attr_name)
            if inspect.ismethod(method) and not attr_name.startswith('_'):
                signature = inspect.signature(method)
                docstring = inspect.getdoc(method)
                attr = dict(signature=signature, docstring=docstring)
                form_class = create_add_form(attr, attr_name, autofill, script, design)
                method_forms[attr_name] = form_class()
        except Exception as e:
            print(f"Error creating form for {attr_name}: {e}")
    return method_forms


def create_form_from_pseudo(pseudo: dict, autofill: bool, script=None, design=True):
    """
    Create forms for pseudo method, used for design routes
    :param pseudo:{'dose_liquid': {
                        "docstring": "some docstring",
                        "signature": Signature(amount_in_ml: float, rate_ml_per_minute: float) }
                    }
    :param autofill:bool if autofill is enabled
    :param script:Script object
    :param design: if design is enabled
    """
    method_forms = {}
    for attr_name, info in pseudo.items():
        # Handle properties (getter/setter)
        if isinstance(info, dict) and info.get('is_property'):
            # Getter
            form_class = create_add_form(info, attr_name, autofill, script, design)
            method_forms[attr_name] = form_class()
            
            # Setter
            if info.get('has_setter'):
                setter_name = f"{attr_name}_(setter)"
                sig = info.get('signature')
                # Infer type from getter return annotation if available
                param_type = inspect._empty
                if sig and sig.return_annotation is not inspect._empty:
                    param_type = sig.return_annotation
                
                # Create a synthetic signature for the setter: (value: type)
                setter_sig = inspect.Signature(
                    parameters=[inspect.Parameter('value', inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=param_type)],
                    return_annotation=None
                )
                
                setter_info = {
                    'signature': setter_sig,
                    'docstring': f"Set {attr_name}"
                }
                form_class_setter = create_add_form(setter_info, setter_name, autofill, script, design)
                method_forms[setter_name] = form_class_setter()
        else:
            # Regular method
            # signature = info.get('signature', {})
            form_class = create_add_form(info, attr_name, autofill, script, design)
            method_forms[attr_name] = form_class()
    return method_forms


def create_form_from_action(action: dict, script=None, design=True):
    '''
    Create forms for single action, used for design routes
    :param action: {'action': 'dose_solid', 'arg_types': {'amount_in_mg': 'float', 'bring_in': 'bool'},
                    'args': {'amount_in_mg': 5.0, 'bring_in': False}, 'id': 9,
                    'instrument': 'deck.sdl', 'return': '', 'uuid': 266929188668995}
    :param script:Script object
    :param design: if design is enabled

    '''

    arg_types = action.get("arg_types", {})
    args = action.get("args", {})
    save_as = action.get("return")
    instrument = action.get("instrument")

    class DynamicForm(FlaskForm):
        pass

    annotation_mapping = {
        "int": (VariableOrIntField if design else IntegerField, 'Enter integer value'),
        "float": (VariableOrFloatField if design else FloatField, 'Enter numeric value'),
        "str": (VariableOrStringField if design else StringField, 'Enter text'),
        "bool": (VariableOrBoolField if design else BooleanField, 'Empty for false')
    }

    # Use explicitly saved order if available, otherwise fallback (e.g. for old actions)
    arg_order = action.get("arg_order", arg_types.keys())

    for name in arg_order:
        param_type = arg_types[name]
        # formatted_param_name = format_name(name)
        value = args.get(name, "")
        if type(value) is dict and value:
            value = next(iter(value))
        if value in (None, "", "None"):
            value = None

        field_kwargs = {
            "label": name,
            "default": value,
            # todo get optional/required from snapshot
            "validators": [],
            "filters": [lambda x: x if x != '' else None],
            **({"script": script})
        }
        if type(param_type) is list:
            none_type = param_type[1]
            if none_type == "NoneType":
                param_type = param_type[0]
        param_type = param_type if type(param_type) is str else f"{param_type}"
        field_class, placeholder_text = annotation_mapping.get(
            param_type,
            (VariableOrStringField if design else StringField, f'Enter {param_type} value')
        )

        if instrument == "math_variable":
            field_class = VariableOrStringField
            placeholder_text = "Enter math expression"
            field_kwargs["validators"] = [InputRequired()]


        render_kwargs = {"placeholder": placeholder_text}

        # Create the field with additional rendering kwargs for placeholder text
        field = field_class(**field_kwargs, render_kw=render_kwargs)
        setattr(DynamicForm, name, field)

    if design:
        if "batch_action" in action:
            batch_action = BooleanField(label='Batch Action', default=bool(action["batch_action"]))
            setattr(DynamicForm, 'batch_action', batch_action)
        return_value = StringField(label='Save value as', default=f"{save_as}", render_kw={"placeholder": "Optional"})
        setattr(DynamicForm, 'return', return_value)
    return DynamicForm()

def create_all_builtin_forms(script):
    all_builtin_forms = {}
    for logic_name in ['if', 'while', 'variable', 'input', 'wait', 'repeat', 'pause', 'math', 'comment']:
        # signature = info.get('signature', {})
        form_class = create_builtin_form(logic_name, script)
        all_builtin_forms[logic_name] = form_class()
    return all_builtin_forms

def create_builtin_form(logic_type, script):
    """
    Create a builtin form {if, while, variable, repeat, wait}
    """
    class BuiltinFunctionForm(FlaskForm):
        pass

    placeholder_text = {
        'wait': 'Enter second',
        'repeat': 'Enter an integer',
        'pause': 'Human Intervention Message',
        'comment': 'Enter comment to log',
        'input': 'Enter prompt message',
        'math': 'Enter math expression, e.g. #x + 5 * (#y - 2)'
    }.get(logic_type, 'Enter statement')
    description_text = {
        'variable': 'Your variable can be numbers, boolean (True or False) or text ("text")',
        'math': "Enter a math expression using #variables, numbers, or returned variables",
    }.get(logic_type, '')
    field_class = {
        'wait': VariableOrFloatField,
        'repeat': VariableOrIntField,
    }.get(logic_type, VariableOrStringField)  # Default to StringField as a fallback
    field_kwargs = {
        "label": f'statement',
        "validators": [InputRequired()] if logic_type in ['wait', "variable", "math"] else [],
        "description": description_text,
        "script": script
    }
    render_kwargs = {"placeholder": placeholder_text}
    field = field_class(**field_kwargs, render_kw=render_kwargs)
    setattr(BuiltinFunctionForm, "statement", field)
    if logic_type == 'variable':
        variable_field = VariableOrStringField(label=f'variable', validators=[InputRequired()],
                                     description="Your variable name cannot include space",
                                     render_kw=render_kwargs, script=script)
        type_field = SelectField(
            'Select Input Type',
            choices=[('int', 'Integer'), ('float', 'Float'), ('str', 'String'), ('bool', 'Boolean')],
            default='str',  # Optional default value
            # coerce = lambda x: None if x == "None" else x
        )
        setattr(BuiltinFunctionForm, "variable", variable_field)
        setattr(BuiltinFunctionForm, "variable_type", type_field)
    elif logic_type == 'input':
        variable_field = VariableOrStringField(label=f'variable', validators=[InputRequired()],
                                     description="Variable to save user input",
                                     render_kw={"placeholder": "Result variable name"}, script=script)
        type_field = SelectField(
            'Select Value Type',
            choices=[('int', 'Integer'), ('float', 'Float'), ('str', 'String'), ('bool', 'Boolean')],
            default='str',
        )
        setattr(BuiltinFunctionForm, "variable", variable_field)
        setattr(BuiltinFunctionForm, "variable_type", type_field)

    elif logic_type == "math":
        math_variable_field = VariableOrStringField(
            label="save_to",
            validators=[InputRequired()],
            description="Variable name to save the result into",
            render_kw={"placeholder": "Result variable name"},
            script=script
        )
        setattr(BuiltinFunctionForm, "math_variable", math_variable_field)

    hidden_field = HiddenField(name=f'builtin_name', render_kw={"value": f'{logic_type}'})
    setattr(BuiltinFunctionForm, "builtin_name", hidden_field)
    return BuiltinFunctionForm


def get_method_from_workflow(function_string):
    """Creates a function from a string and assigns it a new name."""

    namespace = {}
    exec(function_string, globals(), namespace)  # Execute the string in a safe namespace
    func_name = next(iter(namespace))
    # Get the function name dynamically
    return namespace[func_name]


def create_workflow_forms(script, autofill: bool = False, design: bool = False):
    workflow_forms = {}
    functions = {}
    class RegisteredWorkflows:
        pass

    deck_name = script.deck
    workflows = Script.query.filter(Script.deck==deck_name, Script.name != script.name).all()
    for workflow in workflows:
        workflow_name = Script.validate_function_name(workflow.name)
        try:
        # if True:
            compiled_strs = workflow.compile().get('script', "")
            if not compiled_strs:
                continue
            method = get_method_from_workflow(compiled_strs)
            functions[workflow_name] = dict(signature=inspect.signature(method), docstring=inspect.getdoc(method))
            setattr(RegisteredWorkflows, workflow_name, method)

            form_class = create_form_for_method(method, autofill, script, design)

            hidden_method_name = HiddenField(name=f'workflow_name', description=f"{workflow.description}",
                                             render_kw={"value": f'{workflow_name}'})
            if design:
                return_value = StringField(label='Save value as', render_kw={"placeholder": "Optional"})
                setattr(form_class, 'return', return_value)
            setattr(form_class, 'workflow_name', hidden_method_name)
            workflow_forms[workflow_name] = form_class()
        except Exception as e:
            # Log error or skip this workflow
            # print(f"Error loading workflow {workflow_name}: {e}")
            pass
    global_config.registered_workflows = RegisteredWorkflows
    return functions, workflow_forms


def create_action_button(script, stype=None):
    """
    Creates action buttons for design route (design canvas)
    :param script: Script object
    :param stype: script type (script, prep, cleanup)
    """
    stype = stype or script.editing_type
    variables = script.get_variables()
    return [_action_button(i, variables) for i in script.get_script(stype)]


def _action_button(action: dict, variables: dict):
    """
    Creates action button for one action
    :param action: Action dict
    :param variables: created variable dict
    """
    style = {
        "repeat": "background-color: lightsteelblue",
        "if": "background-color: mistyrose",
        "while": "background-color: #a8b5a2",
        "pause": "background-color: palegoldenrod",
        "comment": "background-color: lightgoldenrodyellow",
        "input": "background-color: lightcyan",
    }.get(action['instrument'], "")
    if not style:
        style = "background-color: thistle" if 'batch_action' in action and action["batch_action"] else ""

    if action['instrument'] in ['if', 'while', 'repeat']:
        text = f"{action['action']} {action['args'].get('statement', '')}"
    elif action['instrument'] in ('variable', 'math_variable'):
        text = f"{action['action']} = {action['args'].get('statement')}"
    else:
        # regular action button
        prefix = f"{action['return']} = " if action['return'] else ""
        action_text = f"{action['instrument'].split('.')[-1] if action['instrument'].startswith('deck') else action['instrument']}.{action['action']}"
        arg_string = ""
        if action['args']:
            if type(action['args']) is dict:
                arg_list = []
                if 'arg_order' in action:
                    arg_order = action.get('arg_order')
                else:
                    arg_order = sorted(action['args'].keys())
                for argument_name in arg_order:
                    argument_data = action['args'].get(argument_name)
                    if isinstance(argument_data, dict):
                        if not argument_data:
                            value = argument_data  # Keep the original value if not a dict
                        else:
                            value = next(iter(argument_data))  # Extract the first key if it's a dict
                            # show warning color for variable calling when there is no definition

                            style = "background-color: khaki" if argument_data.get(value) == "function_output" and value not in variables.keys() else ""
                    else:
                        value = argument_data  # Keep the original value if not a dict
                    arg_list.append(f"{argument_name} = {value}")  # Format the key-value pair
                arg_string = "(" + ", ".join(arg_list) + ")"
            else:
                arg_string = f"= {action['args']}"
        text = f"{prefix}{action_text}  {arg_string}"
    return dict(label=text, style=style, uuid=action["uuid"], id=action["id"], instrument=action['instrument'])



from flask import Blueprint, request, session, flash, redirect, url_for, jsonify, render_template, current_app
from flask_login import login_required

from ivoryos.utils import utils
from ivoryos.utils.form import create_form_from_action, create_action_button

steps = Blueprint('design_steps', __name__)


@steps.get("/draft/steps/<int:uuid>")
def get_step(uuid: int):
    """
    .. :quickref: Workflow Design Steps; get an action step editing form

    .. http:get:: /draft/steps/<int:uuid>

    get the editing form for an action step

    :param uuid: The step number id
    :type uuid: int

    :status 200: render template with action step form
    """
    script = utils.get_script_file()
    action = script.find_by_uuid(uuid)
    if action is None:
        return jsonify({"warning": "Step not found, please refresh the page."}), 404

    elif request.method == 'GET':
        forms = create_form_from_action(action, script=script)
        # session['edit_action'] = action
        return render_template("components/edit_action_form.html",
                               action=action,
                               forms=forms)



@steps.post("/draft/steps/<int:uuid>")
def save_step(uuid: int):
    """
    .. :quickref: Workflow Design Steps; save an action step on canvas

    .. http:post:: /draft/steps/<int:uuid>

        save the changes of an action step

        :param uuid: The step number id
        :type uuid: int

    :status 200: render template with action step form
    """
    script = utils.get_script_file()
    action = script.find_by_uuid(uuid)
    warning = None
    if action is not None:
        forms = create_form_from_action(action, script=script)
        kwargs = {field.name: field.data for field in forms if field.name != 'csrf_token'}
        if forms and forms.validate_on_submit():
            save_data = kwargs.pop('return', '')
            # validate return variable name
            save_data = script.validate_function_name(save_data)

            batch_action = kwargs.pop('batch_action', False)

            # literal for args with no typehint
            arg_types = action.get('arg_types', {})
            kwargs = script.validate_variables(kwargs, arg_types)

            script.update_by_uuid(uuid=uuid, args=kwargs, output=save_data, batch_action=batch_action)
        else:
            warning = f"Compilation failed: {str(forms.errors)}"
    utils.post_script_file(script)
    try:
        exec_string = script.compile(current_app.config['SCRIPT_FOLDER'])
    except Exception as e:
        exec_string = {}
        warning = f"Compilation failed: {str(e)}"
    session['python_code'] = exec_string
    design_buttons = {stype: create_action_button(script, stype) for stype in script.stypes}
    return render_template("components/canvas_main.html",
                           script=script,
                           buttons_dict=design_buttons,
                           warning=warning)

@steps.delete("/draft/steps/<int:uuid>")
def delete_step(uuid: int):
    """
    .. :quickref: Workflow Design Steps; delete an action step on canvas

    .. http:delete:: /draft/steps/<int:uuid>

        delete an action step

        :param uuid: The step number id
        :type uuid: int

    :status 200: render template with action step form
    """
    script = utils.get_script_file()
    if request.method == 'DELETE':
        script.delete_action(uuid)
    utils.post_script_file(script)
    warning = None
    try:
        exec_string = script.compile(current_app.config['SCRIPT_FOLDER'])
    except Exception as e:
        exec_string = {}
        warning = f"Compilation failed: {str(e)}"
    session['python_code'] = exec_string
    design_buttons = {stype: create_action_button(script, stype) for stype in script.stypes}
    return render_template("components/canvas_main.html",
                               script=script,
                               buttons_dict=design_buttons, warning=warning)


@steps.route("/draft/steps/<int:uuid>/duplicate", methods=["POST"], strict_slashes=False,)
def duplicate_action(uuid: int):
    """
    .. :quickref: Workflow Design Steps; duplicate an action step on canvas

    .. http:post:: /draft/steps/<int:uuid>/duplicate

    :param uuid: The step number uuid
    :type uuid: int

    :status 200: render new design script template
    """

    # back = request.referrer
    script = utils.get_script_file()
    script.duplicate_action(uuid)
    utils.post_script_file(script)
    warning = None
    try:
        exec_string = script.compile(current_app.config['SCRIPT_FOLDER'])
    except Exception as e:
        exec_string = {}
        warning = f"Compilation failed: {str(e)}"
    session['python_code'] = exec_string
    design_buttons = {stype: create_action_button(script, stype) for stype in script.stypes}

    return render_template("components/canvas_main.html",
                         script=script,
                         buttons_dict=design_buttons, warning=warning)


@steps.route("/draft/steps/order", methods=['POST'])
@login_required
def update_list():
    """
    .. :quickref: Workflow Design Steps; update the order of steps in the design canvas when reordering steps.

    .. http:post:: /draft/steps/order

    Update the order of steps in the design canvas when reordering steps.

    :form order: A comma-separated string representing the new order of steps.
    :status 200: Successfully updated the order of steps.
    """
    order = request.form['order']
    script = utils.get_script_file()
    script.currently_editing_order = order.split(",", len(script.currently_editing_script))
    script.sort_actions()
    warning = None

    utils.post_script_file(script)
    try:
        exec_string = script.compile(current_app.config['SCRIPT_FOLDER'])
    except Exception as e:
        exec_string = {}
        warning = f"Compilation failed: {str(e)}"
    session['python_code'] = exec_string

    # Return the updated canvas HTML instead of JSON
    design_buttons = {stype: create_action_button(script, stype) for stype in script.stypes}
    return render_template("components/canvas_main.html",
                           script=script,
                           buttons_dict=design_buttons, warning=warning)


from datetime import datetime

from ivoryos.utils.db_models import Script, WorkflowRun, WorkflowStep
from ivoryos import db

def test_database_scripts_page(auth):
    """
    GIVEN an authenticated user
    WHEN they access the script database page
    THEN the page should load and show their scripts
    """
    # First, create a script so the page has something to render
    with auth.application.app_context():
        script = Script(name='test_script', author='testuser')
        db.session.add(script)
        db.session.commit()

    response = auth.get('/ivoryos/database/scripts/', follow_redirects=True)
    assert response.status_code == 200
    # assert b'Scripts Database' in response.data
    assert b'<title>IvoryOS | Design Database</title>' in response.data

def test_database_workflows_page(auth):
    """
    GIVEN an authenticated user
    WHEN they access the workflow database page
    THEN the page should load and show past workflow runs
    """
    # Create a workflow run to display
    with auth.application.app_context():
        run = WorkflowRun(name="untitled", platform="deck",start_time=datetime.now())
        db.session.add(run)
        db.session.commit()
        run_id = run.id

    response = auth.get('/ivoryos/database/workflows/', follow_redirects=True)
    assert response.status_code == 200
    assert b'Workflow ID' in response.data
    # assert b'run_id' in response.data

def test_view_specific_workflow(auth):
    """
    GIVEN an authenticated user and an existing workflow run
    WHEN they access the specific URL for that workflow
    THEN the detailed view for that run should be displayed
    """
    with auth.application.app_context():
        run = WorkflowRun(name='test_workflow', platform='test_platform', start_time=datetime.now())
        db.session.add(run)
        db.session.commit()
        run_id = run.id

        step = WorkflowStep(method_name='test_step', workflow_id=run_id, phase="main", run_error=False, start_time=datetime.now())
        db.session.add(step)
        db.session.commit()
        # run_id = run.id

    response = auth.get(f'/ivoryos/database/workflows/{run_id}', follow_redirects=True)
    assert response.status_code == 200
    # assert b'test_step' in response.data # Check for a title on the view page
    assert b'test_workflow' in response.data


import ast
import importlib
import inspect
import json
import logging
import os
import pickle
import socket
import subprocess
import sys
from collections import Counter
from enum import Enum

import flask
from flask import session
from flask_login import current_user
from flask_socketio import SocketIO

from ivoryos.utils.db_models import Script
from ivoryos.utils.decorators import BUILDING_BLOCKS

def get_script_file():
    """Get script from Flask session and returns the script"""
    session_script = session.get("scripts")
    if session_script:
        s = Script()
        s.__dict__.update(**session_script)
        return s
    else:
        return Script(author=current_user.get_id(),)


def post_script_file(script, is_dict=False):
    """
    Post script to Flask. Script will be converted to a dict if it is a Script object
    :param script: Script to post
    :param is_dict: if the script is a dictionary,
    """
    if is_dict:
        session['scripts'] = script
    else:
        session['scripts'] = script.as_dict()


def create_gui_dir(parent_path):
    """
    Creates folders for ivoryos data
    """
    os.makedirs(parent_path, exist_ok=True)
    for path in ["config_csv", "scripts", "results", "pseudo_deck"]:
        os.makedirs(os.path.join(parent_path, path), exist_ok=True)


def save_to_history(filepath, history_path):
    """
    For manual deck connection only
    save deck file path that successfully connected to ivoryos to a history file
    """
    connections = []
    try:
        with open(history_path, 'r') as file:
            lines = file.read()
            connections = lines.split('\n')
    except FileNotFoundError:
        pass
    if filepath not in connections:
        with open(history_path, 'a') as file:
            file.writelines(f"{filepath}\n")


def import_history(history_path):
    """
    For manual deck connection only
    load deck connection history from history file
    """
    connections = []
    try:
        with open(history_path, 'r') as file:
            lines = file.read()
            connections = lines.split('\n')
    except FileNotFoundError:
        pass
    connections = [i for i in connections if not i == '']
    return connections


def available_pseudo_deck(path):
    """
    load pseudo deck (snapshot) from connection history
    """
    return os.listdir(path)


def _inspect_class(class_object=None, debug=False):
    """
    inspect class object: inspect function signature if not name.startswith("_")
    :param class_object: class object
    :param debug: debug mode will inspect function.startswith("_")
    :return: function: Dict[str, Dict[str, Union[Signature, str, None]]]
    """
    functions = {}
    under_score = "_"
    if debug:
        under_score = "__"

    for function, method in inspect.getmembers(type(class_object), predicate=callable):
        if not function.startswith(under_score) and not function.isupper():
            try:
                annotation = inspect.signature(method)
                docstring = inspect.getdoc(method)
                coroutine = inspect.iscoroutinefunction(method)
                functions[function] = dict(signature=annotation, docstring=docstring, coroutine=coroutine,)

            except Exception:
                pass

    for function, prop in inspect.getmembers(type(class_object), lambda x: isinstance(x, property)):
        if not function.startswith(under_score) and not function.isupper():
            try:
                annotation = inspect.signature(prop.fget) if prop.fget else None
                docstring = inspect.getdoc(prop)
                functions[function] = dict(signature=annotation, docstring=docstring, coroutine=False,
                                           is_property=True, has_setter=prop.fset is not None)
            except Exception:
                pass
    return functions


def _get_type_from_parameters(arg, parameters):
    """get argument types from inspection"""
    # TODO
    arg_type = ''
    try:
        if isinstance(parameters, inspect.Signature):
            if arg not in parameters.parameters:
                return arg_type
            annotation = parameters.parameters[arg].annotation
        elif isinstance(parameters, dict):
            annotation = parameters.get(arg, '')
        else:
            annotation = ''
    except Exception:
        return arg_type

    if isinstance(annotation, str):
        arg_type = annotation
    elif annotation is not inspect._empty:
        if annotation.__module__ == 'typing':

            if hasattr(annotation, '__origin__'):
                origin = annotation.__origin__
                if hasattr(origin, '_name') and origin._name in ["Optional", "Union"]:
                    arg_type = [i.__name__ for i in annotation.__args__]
                elif hasattr(origin, '__name__'):
                    arg_type = origin.__name__
                # todo other types
        elif annotation.__module__ == 'types':
            arg_type = [i.__name__ for i in annotation.__args__]

        else:
            arg_type = annotation.__name__
    return arg_type


def _convert_by_str(args, arg_types):
    """
    Converts a value to type through eval(f'{type}("{args}")')
    v1.3.4 TODO try str lastly, otherwise it's always converted to str
    """
    if type(arg_types) is not list:
        arg_types = [arg_types]
    for arg_type in arg_types:
        if not arg_type == "any":
            try:
                args = eval(f'{arg_type}("{args}")') if type(args) is str else eval(f'{arg_type}({args})')
                return args
            except Exception:
                raise TypeError(f"Input type error: cannot convert '{args}' to {arg_type}.")
    return args


def _convert_by_class(args, arg_types):
    """
    Converts a value to type through type(arg)
    """
    if arg_types.__module__ == 'builtins':
        args = arg_types(args)
        return args
    elif arg_types.__module__ == "typing":
        for i in arg_types.__args__:  # for typing.Union
            try:
                args = i(args)
                return args
            except Exception:
                pass
        raise TypeError("Input type error.")
    # else:
    #     args = globals()[args]
    return args


def convert_config_type(args, arg_types, is_class: bool = False):
    """
    Converts an argument from str to an arg type
    """
    if args:
        Script.eval_list(args, arg_types)
        for arg in args:
            if arg not in arg_types.keys():
                raise ValueError("config file format not supported.")
            if args[arg] == '' or args[arg] == "None":
                args[arg] = None
            # elif args[arg] == "True" or args[arg] == "False":
            #     args[arg] = bool_dict[args[arg]]
            else:
                arg_type = arg_types[arg]
                try:
                    args[arg] = ast.literal_eval(args[arg])
                except Exception:
                    pass
                if type(args[arg]) is not arg_type and not type(args[arg]).__name__ == arg_type:
                    if is_class:
                        # if arg_type.__module__ == 'builtins':
                        args[arg] = _convert_by_class(args[arg], arg_type)
                    else:
                        args[arg] = _convert_by_str(args[arg], arg_type)
    return args


def import_module_by_filepath(filepath: str, name: str):
    """
    Import module by file path
    :param filepath: full path of module
    :param name: module's name
    """
    spec = importlib.util.spec_from_file_location(name, filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


class SocketIOHandler(logging.Handler):
    def __init__(self, socketio: SocketIO):
        super().__init__()
        self.formatter = logging.Formatter('%(asctime)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        self.socketio = socketio

    def emit(self, record):
        message = self.format(record)
        # session["last_log"] = message
        self.socketio.emit('log', {'message': message})


def start_logger(socketio: SocketIO, logger_name: str, log_filename: str = None):
    """
    stream logger to web through web socketIO
    """
    # logging.basicConfig( format='%(asctime)s - %(message)s')
    formatter = logging.Formatter(fmt='%(asctime)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)
    file_handler = logging.FileHandler(filename=log_filename, )
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    # console_logger = logging.StreamHandler()  # stream to console
    # logger.addHandler(console_logger)
    socketio_handler = SocketIOHandler(socketio)
    logger.addHandler(socketio_handler)
    return logger


def get_arg_type(args, parameters):
    """get argument type from signature"""
    arg_types = {}
    # print(args, parameters)
    parameters = parameters.get("signature")
    if args:
        for arg in args:
            arg_types[arg] = _get_type_from_parameters(arg, parameters)
    return arg_types


def install_and_import(package, package_name=None):
    """
    Install the package and import it
    :param package: package to import and install
    :param package_name: pip install package name if different from package
    """
    try:
        # Check if the package is already installed
        importlib.import_module(package)
        # print(f"{package} is already installed.")
    except ImportError:
        # If not installed, install it
        # print(f"{package} is not installed. Installing now...")
        subprocess.check_call([sys.executable, "-m", "pip", "install", package_name or package])
        # print(f"{package} has been installed successfully.")


def web_config_entry_wrapper(data: dict, config_type: list):
    """
    Wrap the data dictionary from web config entries during execution configuration
    :param data: data dictionary
    :param config_type: data entry types ["str", "int", "float", "bool"]
    """
    rows = {}  # Dictionary to hold webui_data organized by rows

    # Organize webui_data by rows
    for key, value in data.items():
        if value:  # Only process non-empty values
            # Extract the field name and row index
            field_name, row_index = key.split('[')
            row_index = int(row_index.rstrip(']'))

            # If row not in rows, create a new dictionary for that row
            if row_index not in rows:
                rows[row_index] = {}

            # Add or update the field value in the specific row's dictionary
            rows[row_index][field_name] = value

    # Filter out any empty rows and create a list of dictionaries
    filtered_rows = [row for row in rows.values() if len(row) == len(config_type)]

    return filtered_rows


def create_deck_snapshot(deck, save: bool = False, output_path: str = '', exclude_names: list = []):
    """
    Create a deck snapshot of the given script
    :param deck: python module name to create the deck snapshot from e.g. __main__
    :param save: save the deck snapshot into pickle file
    :param output_path: path to save the pickle file
    :param exclude_names: module names to exclude from deck snapshot
    """
    exclude_classes = (flask.Blueprint, logging.Logger)

    deck_snapshot = {}
    included = {}
    excluded = {}
    failed = {}

    for name, val in vars(deck).items():
        qualified_name = f"deck.{name}"

        # Exclusion checks
        if (
                type(val).__module__ == 'builtins'
                or name[0].isupper()
                or name.startswith("_")
                or isinstance(val, exclude_classes)
                or name in exclude_names
        ):
            excluded[qualified_name] = type(val).__name__
            continue

        try:
            deck_snapshot[qualified_name] = _inspect_class(val)
            included[qualified_name] = type(val).__name__
        except Exception as e:
            failed[qualified_name] = str(e)

    # Final result
    deck_summary = {
        "included": included,
        # "excluded": excluded,
        "failed": failed
    }

    def print_deck_snapshot(deck_summary):
        def print_section(title, items):
            print(f"\n=== {title} ({len(items)}) ===")
            if not items:
                return
            for name, class_type in items.items():
                print(f"  {name}: {class_type}")

        print_section("✅ INCLUDED MODULES", deck_summary["included"])
        print_section("❌ FAILED MODULES", deck_summary["failed"])
        print("\n")

    print_deck_snapshot(deck_summary)

    if deck_snapshot and save:
        # pseudo_deck = parse_dict
        parse_dict = deck_snapshot.copy()
        parse_dict["deck_name"] = os.path.splitext(os.path.basename(deck.__file__))[
            0] if deck.__name__ == "__main__" else deck.__name__
        with open(os.path.join(output_path, f"{parse_dict['deck_name']}.pkl"), 'wb') as file:
            pickle.dump(parse_dict, file)
    return deck_snapshot


def create_block_snapshot(save: bool = False, output_path: str = ''):
    block_snapshot = {}
    included = {}
    failed = {}
    for category, data in BUILDING_BLOCKS.items():
        key = f"blocks.{category}"
        block_snapshot[key] = {}

        for func_name, meta in data.items():
            func = meta["func"]
            block_snapshot[key][func_name] = {
                "signature": meta["signature"],
                "docstring": meta["docstring"],
                "coroutine": meta["coroutine"],
                "path": f"{func.__module__}.{func.__qualname__}"
            }
    if block_snapshot:
        print(f"\n=== ✅ BUILDING_BLOCKS ({len(block_snapshot)}) ===")
        for category, blocks in block_snapshot.items():
            print(f"  {category}: ", ",".join(blocks.keys()))
    return block_snapshot

def load_deck(pkl_name: str):
    """
    Loads a pickled deck snapshot from disk on offline mode
    :param pkl_name: name of the pickle file
    """
    if not pkl_name:
        return None
    try:
        with open(pkl_name, 'rb') as f:
            pseudo_deck = pickle.load(f)
        return pseudo_deck
    except FileNotFoundError:
        return None


def check_config_duplicate(config):
    """
    Checks if the config entry has any duplicate
    :param config: [{"arg": 1}, {"arg": 1}, {"arg": 1}]
    :return: [True, False]
    """
    hashable_data = [tuple(sorted(d.items())) for d in config]
    return any(count > 1 for count in Counter(hashable_data).values())


def get_method_from_workflow(function_string, func_name="workflow"):
    """Creates a function from a string and assigns it a new name."""

    namespace = {}
    exec(function_string, globals(), namespace)  # Execute the string in a safe namespace
    # func_name = next(iter(namespace))
    # Get the function name dynamically
    return namespace[func_name]

# def load_workflows(script):
#
#     class RegisteredWorkflows:
#         pass
#     deck_name = script.deck
#     workflows = Script.query.filter(Script.deck == deck_name, Script.name != script.name, Script.registered==True).all()
#     for workflow in workflows:
#         compiled_strs = workflow.compile().get('script', "")
#         method = get_method_from_workflow(compiled_strs, func_name=workflow.name)
#         setattr(RegisteredWorkflows, workflow.name, staticmethod(method))
#     global_config.registered_workflows = RegisteredWorkflows()


def get_local_ip():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('10.255.255.255', 1))  # Dummy address to get interface IP
        ip = s.getsockname()[0]
    except Exception:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


def safe_dump(obj):
    try:
        json.dumps(obj)
        return obj
    except (TypeError, OverflowError):
        if isinstance(obj, Enum):
            return obj.value
        return repr(obj)  # store readable representation


def create_module_snapshot(module):
    classes = inspect.getmembers(module, inspect.isclass)
    api_variables = {}
    for i in classes:
        # globals()[i[0]] = i[1]
        api_variables[i[0]] = i[1]
    return api_variables


import os
from flask import Blueprint,  request,current_app, send_file
from flask_login import login_required

from ivoryos.utils.client_proxy import ProxyGenerator
from ivoryos.utils.global_config import GlobalConfig

global_config = GlobalConfig()

control_file = Blueprint('file', __name__)



@control_file.route("/files/proxy", strict_slashes=False)
@login_required
def download_proxy():
    """
    .. :quickref: Direct Control Files; Download proxy Python interface

    download proxy Python interface

    .. http:get:: /files/proxy
    """
    generator = ProxyGenerator(request.url_root)
    snapshot = global_config.deck_snapshot.copy()

    filepath = generator.generate_from_flask_route(
        snapshot,
        request.url_root,
        current_app.config["OUTPUT_FOLDER"]
    )

    return send_file(os.path.abspath(filepath), as_attachment=True)


from flask import session

from ivoryos.utils.global_config import GlobalConfig


global_config = GlobalConfig()


def find_instrument_by_name(name: str):
    """
    find instrument class object by instance name
    """
    if name.startswith("deck"):
        name = name.replace("deck.", "")
        return getattr(global_config.deck, name)
    elif name.startswith("blocks"):
        return global_config.building_blocks[name]
    elif name in global_config.defined_variables:
        return global_config.defined_variables[name]
    elif name in globals():
        return globals()[name]


def get_session_by_instrument(session_name, instrument):
    """get data from session by instrument"""
    session_object = session.get(session_name, {})
    functions = session_object.get(instrument, [])
    return functions


def post_session_by_instrument(session_name, instrument, data):
    """
    save new data to session by instrument
    :param session_name: "card_order" or "hidden_functions"
    :param instrument: function name of class object
    :param data: order list or hidden function list
    """
    session_object = session.get(session_name, {})
    session_object[instrument] = data
    session[session_name] = session_object



from flask import Blueprint, redirect, url_for, flash, request, render_template, session, current_app, jsonify
from flask_login import login_required, current_user

from ivoryos.utils.db_models import Script, db, WorkflowRun, WorkflowStep
from ivoryos.utils.utils import get_script_file, post_script_file

library = Blueprint('library', __name__, template_folder='templates')



@library.route("/<string:script_name>", methods=["GET", "POST", "DELETE"])
@login_required
def workflow_script(script_name:str):
    # todo: split this into two routes, one for GET and POST, another for DELETE
    """
    .. :quickref: Workflow Script Database; get, post, delete a workflow script

    .. http:get:: /library/<string: script_name>

    :param script_name: script name
    :type script_name: str
    :status 302: redirect to :http:get:`/ivoryos/draft`

    .. http:post:: /library/<string: script_name>

    :param script_name: script name
    :type script_name: str
    :status 200: json response with success status

    .. http:delete:: /library/<string: script_name>

    :param script_name: script name
    :type script_name: str
    :status 302: redirect to :http:get:`/ivoryos/draft`

    """
    row = Script.query.get(script_name)
    if request.method == "DELETE":
        if not row:
            return jsonify(success=False)
        db.session.delete(row)
        db.session.commit()
        return jsonify(success=True)
    if request.method == "GET":
        if not row:
            return jsonify(success=False)
        script = Script(**row.as_dict())
        post_script_file(script)
        pseudo_name = session.get("pseudo_deck", "")
        off_line = current_app.config["OFF_LINE"]
        if off_line and pseudo_name and not script.deck == pseudo_name:
            flash(f"Choose the deck with name {script.deck}")
        if request.accept_mimetypes.best_match(['application/json', 'text/html']) == 'application/json':
            return jsonify({
                "script": script.as_dict(),
                "python_script": script.compile(),
            })
        return redirect(url_for('design.experiment_builder'))
    if request.method == "POST":
        status = publish()
        return jsonify(status)
    return None


def publish():
    script = get_script_file()

    if script.author is None:
        script.author = current_user.get_id()
    if not script.name or not script.deck:
        return {"success": False, "error": "Deck cannot be empty, try to re-submit deck configuration on the left panel"}
    row = Script.query.get(script.name)
    if row and row.status == "finalized":
        return {"success": False, "error": "This is a protected script, use save as to rename."}

    elif row and current_user.get_id() != row.author:
        return {"success": False, "error": "You are not the author, use save as to rename."}
    else:
        db.session.merge(script)
        db.session.commit()
        return {"success": True, "message": "Script published successfully"}


@library.get("/", strict_slashes=False)
@login_required
def load_from_database():
    """
    .. :quickref: Script Database; database page

    backend control through http requests

    .. http:get:: /library


    """
    session.pop('edit_action', None)  # reset cache
    query = Script.query
    search_term = request.args.get("keyword", None)
    deck_name = request.args.get("deck", None)
    if search_term:
        query = query.filter(Script.name.like(f'%{search_term}%'))
    if deck_name is None:
        temp = Script.query.with_entities(Script.deck).distinct().all()
        deck_list = [i[0] for i in temp]
    else:
        query = query.filter(Script.deck == deck_name)
        deck_list = ["ALL"]
    page = request.args.get('page', default=1, type=int)
    per_page = 10

    scripts = query.paginate(page=page, per_page=per_page, error_out=False)
    if request.accept_mimetypes.best_match(['application/json', 'text/html']) == 'application/json':
        scripts = query.all()
        script_names = [script.name for script in scripts]
        return jsonify({
            "workflows": script_names,
        })
    else:
        # return HTML
        return render_template("library.html", scripts=scripts, deck_list=deck_list, deck_name=deck_name)




@library.post("/", strict_slashes=False)
@login_required
def save_as():
    """
    .. :quickref: Script Database; save the script as

    save the current workflow script as

    .. http:post:: /library

    : form run_name: new workflow name
    :status 302: redirect to :http:get:`/ivoryos/draft`

    """
    if request.method == "POST":
        run_name = request.form.get("run_name")
        description = request.form.get("description")
        ## TODO: check if run_name is valid
        register_workflow = request.form.get("register_workflow")
        script = get_script_file()
        script.save_as(run_name)
        script.registered = register_workflow == "on"
        script.author = current_user.get_id()
        script.description = description
        post_script_file(script)
        status = publish()
        if request.accept_mimetypes.best_match(['application/json', 'text/html']) == 'application/json':
            return jsonify(status)
        else:
            if status["success"]:
                flash("Script saved successfully")
            else:
                flash(status["error"], "error")
            return redirect(url_for('design.experiment_builder'))




# optimizers/registry.py

from ivoryos.optimizer.ax_optimizer import AxOptimizer
from ivoryos.optimizer.baybe_optimizer import BaybeOptimizer
from ivoryos.optimizer.nimo_optimizer import NIMOOptimizer

OPTIMIZER_REGISTRY = {
    "ax": AxOptimizer,
    "baybe": BaybeOptimizer,
    "nimo": NIMOOptimizer,
}



import ast
try:
    from ast import unparse as ast_unparse
except ImportError:
    import astor
    ast_unparse = astor.to_source

import builtins
import json
import keyword
import re
import uuid
from datetime import datetime
from typing import Dict

from flask_login import UserMixin
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy_utils import JSONType

db = SQLAlchemy()


class User(db.Model, UserMixin):
    __tablename__ = 'user'
    # id = db.Column(db.Integer)
    username = db.Column(db.String(50), primary_key=True, unique=True, nullable=False)
    # email = db.Column(db.String)
    hashPassword = db.Column(db.String(255))

    # New columns for logo customization
    settings = db.Column(JSONType, nullable=True)

    # password = db.Column()
    def __init__(self, username, password):
        # self.id = id
        self.username = username
        # self.email = email
        self.hashPassword = password

    def get_id(self):
        return self.username


class Script(db.Model):
    __tablename__ = 'script'
    # id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(50), primary_key=True, unique=True)
    deck = db.Column(db.String(50), nullable=True)
    status = db.Column(db.String(50), nullable=True)
    script_dict = db.Column(JSONType, nullable=True)
    time_created = db.Column(db.String(50), nullable=True)
    last_modified = db.Column(db.String(50), nullable=True)
    id_order = db.Column(JSONType, nullable=True)
    editing_type = db.Column(db.String(50), nullable=True)
    author = db.Column(db.String(50), nullable=False)
    description = db.Column(db.String(255), nullable=True)
    # registered = db.Column(db.Boolean, nullable=True, default=False)

    def __init__(self, name=None, deck=None, status=None, script_dict: dict = None, id_order: dict = None,
                 time_created=None, last_modified=None, editing_type=None, author: str = None,
                 # registered:bool=False,
                 description: str = None,
                 python_script: str = None
                 ):
        if script_dict is None:
            script_dict = {"prep": [], "script": [], "cleanup": []}
        elif type(script_dict) is not dict:
            script_dict = json.loads(script_dict)
        if id_order is None:
            id_order = {"prep": [], "script": [], "cleanup": []}
        elif type(id_order) is not dict:
            id_order = json.loads(id_order)
        if status is None:
            status = 'editing'
        if time_created is None:
            time_created = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        if last_modified is None:
            last_modified = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        if editing_type is None:
            editing_type = "script"
        if description is None:
            description = ""
        self.name = name
        self.deck = deck
        self.status = status
        self.script_dict = script_dict
        self.time_created = time_created
        self.last_modified = last_modified
        self.id_order = id_order
        self.editing_type = editing_type
        self.author = author
        self.python_script = python_script
        self.description = description
        # self.r = registered

    def as_dict(self):
        data = dict(self.__dict__)  # shallow copy
        data.pop('_sa_instance_state', None)
        return data

    def get(self):
        workflows = db.session.query(Script).all()
        # result = script_schema.dump(workflows)
        return workflows

    def find_by_uuid(self, uuid):
        for stype in self.script_dict:
            for action in self.script_dict[stype]:

                if action['uuid'] == int(uuid):
                    return action

    def _convert_type(self, args, arg_types):
        if arg_types in ["list", "tuple", "set"]:
            try:
                args = ast.literal_eval(args)
                return args
            except Exception:
                pass
        if type(arg_types) is not list:
            arg_types = [arg_types]
        for arg_type in arg_types:
            try:
                # print(arg_type)
                args = eval(f"{arg_type}('{args}')")
                return
            except Exception:

                pass
        raise TypeError(f"Input type error: cannot convert '{args}' to {arg_type}.")

    def update_by_uuid(self, uuid, args, output, batch_action=False):
        action = self.find_by_uuid(uuid)
        if not action:
            return
        arg_types = action['arg_types']
        if type(action['args']) is dict:
            # pass
            self.eval_list(args, arg_types)
        else:
            pass
        action['args'] = args
        action['return'] = output
        action['batch_action'] = batch_action

    @staticmethod
    def eval_list(args, arg_types):
        for arg in args:
            arg_type = arg_types[arg]
            if arg_type in ["list", "tuple", "set"]:

                if type(arg) is str and not args[arg].startswith("#"):
                    # arg_types = arg_types[arg]
                    # if arg_types in ["list", "tuple", "set"]:
                    convert_type = getattr(builtins, arg_type)  # Handle unknown types s
                    try:
                        output = ast.literal_eval(args[arg])
                        if type(output) not in [list, tuple, set]:
                            output = [output]
                        args[arg] = convert_type(output)
                        # return args
                    except ValueError:
                        _list = ''.join(args[arg]).split(',')
                        # convert_type = getattr(builtins, arg_types)  # Handle unknown types s
                        args[arg] = convert_type([s.strip() for s in _list])

    @property
    def stypes(self):
        return list(self.script_dict.keys())

    @property
    def currently_editing_script(self):
        return self.script_dict[self.editing_type]

    @currently_editing_script.setter
    def currently_editing_script(self, script):
        self.script_dict[self.editing_type] = script

    @property
    def currently_editing_order(self):
        return self.id_order[self.editing_type]

    @currently_editing_order.setter
    def currently_editing_order(self, script):
        self.id_order[self.editing_type] = script

    def update_time_stamp(self):
        self.last_modified = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    def get_script(self, stype: str):
        return self.script_dict[stype]

    def isEmpty(self) -> bool:
        if not (self.script_dict['script'] or self.script_dict['prep'] or self.script_dict['cleanup']):
            return True
        return False

    def _sort(self, script_type):
        if len(self.id_order[script_type]) > 0:
            for action in self.script_dict[script_type]:
                for i in range(len(self.id_order[script_type])):
                    if action['id'] == int(self.id_order[script_type][i]):
                        # print(i+1)
                        action['id'] = i + 1
                        break
            self.id_order[script_type].sort(key=int)
            if not int(self.id_order[script_type][-1]) == len(self.script_dict[script_type]):
                new_order = list(range(1, len(self.script_dict[script_type]) + 1))
                self.id_order[script_type] = [str(i) for i in new_order]
            self.script_dict[script_type].sort(key=lambda x: int(x['id']))

    def sort_actions(self, script_type=None):
        if script_type:
            self._sort(script_type)
        else:
            for i in self.stypes:
                self._sort(i)

    def add_action(self, action: dict, insert_position=None):
        current_len = len(self.currently_editing_script)
        action_to_add = action.copy()
        action_to_add['id'] = current_len + 1
        action_to_add['uuid'] = uuid.uuid4().fields[-1]
        self.currently_editing_script.append(action_to_add)
        self._insert_action(insert_position, current_len)
        self.update_time_stamp()

    def add_variable(self, statement, variable, variable_type, insert_position=None):
        variable = self.validate_function_name(variable)
        convert_type = getattr(builtins, variable_type)
        if isinstance(statement, str) and statement.startswith("#"):
            pass
        # boolean values
        elif variable_type == "bool":
            statement = True if statement.lower() in ["true", "y", "t", "yes"] else False
        else:
            statement = convert_type(statement)
        current_len = len(self.currently_editing_script)
        uid = uuid.uuid4().fields[-1]
        action = {"id": current_len + 1, "instrument": 'variable', "action": variable,
                        "args": {"statement": 'None' if statement == '' else statement}, "return": '', "uuid": uid,
                        "arg_types": {"statement": variable_type}}
        self.currently_editing_script.append(action)
        self._insert_action(insert_position, current_len)
        self.update_time_stamp()

    def add_math_variable(self, statement, math_variable, insert_position=None):
        math_variable = self.validate_function_name(math_variable)

        current_len = len(self.currently_editing_script)
        uid = uuid.uuid4().fields[-1]
        action = {"id": current_len + 1, "instrument": 'math_variable', "action": math_variable,
                        "args": {"statement": 'None' if statement == '' else statement}, "return": '', "uuid": uid,
                        "arg_types": {"statement": 'float'}}
        self.currently_editing_script.append(action)
        self._insert_action(insert_position, current_len)
        self.update_time_stamp()

    def add_input_action(self, statement, variable, variable_type, insert_position=None):
        current_len = len(self.currently_editing_script)
        uid = uuid.uuid4().fields[-1]
        action = {"id": current_len + 1, "instrument": 'input', "action": variable,
                  "args": {"statement": statement, "variable": variable}, "return": variable, "uuid": uid,
                  "arg_types": {"statement": variable_type}}
        self.currently_editing_script.append(action)
        self._insert_action(insert_position, current_len)
        self.update_time_stamp()

    def _insert_action(self, insert_position, current_len, action_len:int=1):
        if not len(self.currently_editing_order) == current_len:
            # check if order exists, if not, create a new one
            self.currently_editing_order = list(range(1, current_len + action_len + 1))

        if insert_position is None:
            self.currently_editing_order.extend([str(current_len + i + 1) for i in range(action_len)])
        else:
            index = int(insert_position) - 1
            self.currently_editing_order[index:index] = [str(current_len + i + 1) for i in range(action_len)]
            self.sort_actions()

    def get_added_variables(self):
        added_variables: Dict[str, str] = {action["action"]: action["arg_types"]["statement"] for action in
                                           self.currently_editing_script if action["instrument"] == "variable"}
        added_math_variables: Dict[str, str] = {action["action"]: action["arg_types"]["statement"] for action in
                                                self.currently_editing_script if action["instrument"] == "math_variable"}
        all_added_variables = {**added_variables, **added_math_variables}
        return all_added_variables

    def get_output_variables(self):
        output_variables: Dict[str, str] = {action["return"]: "function_output" for action in
                                            self.currently_editing_script if action["return"]}

        return output_variables

    def get_variables(self):
        output_variables: Dict[str, str] = self.get_output_variables()
        added_variables = self.get_added_variables()
        output_variables.update(added_variables)

        return output_variables

    def validate_variables(self, kwargs, arg_types: dict = None):
        """
        Validates the kwargs passed to the Script
        """
        output_variables: Dict[str, str] = self.get_variables()
        # print(output_variables)
        for key, value in kwargs.items():
            if isinstance(value, str):
                if value in output_variables:
                    var_type = output_variables[value]
                    kwargs[key] = f"#{value}"
                elif value.startswith("#"):
                    kwargs[key] = f"#{self.validate_function_name(value[1:])}"
                else:
                    # attempt to convert to numerical or bool value for args with no type hint
                    type_hint = arg_types.get(key, "") if arg_types else ""
                    if not type_hint:
                        try:
                            converted = ast.literal_eval(value)
                            if isinstance(converted, (int, float, bool)):
                                kwargs[key] = converted
                        except (ValueError, SyntaxError):
                            pass
        return kwargs

    def add_logic_action(self, logic_type: str, statement, insert_position=None):
        current_len = len(self.currently_editing_script)
        uid = uuid.uuid4().fields[-1]
        logic_dict = {
            "if":
                [
                    {"id": current_len + 1, "instrument": 'if', "action": 'if',
                     "args": {"statement": 'True' if statement == '' else statement},
                     "return": '', "uuid": uid, "arg_types": {"statement": ''}},
                    {"id": current_len + 2, "instrument": 'if', "action": 'else', "args": {}, "return": '',
                     "uuid": uid},
                    {"id": current_len + 3, "instrument": 'if', "action": 'endif', "args": {}, "return": '',
                     "uuid": uid},
                ],
            "while":
                [
                    {"id": current_len + 1, "instrument": 'while', "action": 'while',
                     "args": {"statement": 'False' if statement == '' else statement}, "return": '', "uuid": uid,
                     "arg_types": {"statement": ''}},
                    {"id": current_len + 2, "instrument": 'while', "action": 'endwhile', "args": {}, "return": '',
                     "uuid": uid},
                ],

            "wait":
                [
                    {"id": current_len + 1, "instrument": 'wait', "action": "wait",
                     "args": {"statement": 1 if statement == '' else statement},
                     "return": '', "uuid": uid, "arg_types": {"statement": "float"}},
                ],
            "repeat":
                [
                    {"id": current_len + 1, "instrument": 'repeat', "action": "repeat",
                     "args": {"statement": 1 if statement == '' else statement}, "return": '', "uuid": uid,
                     "arg_types": {"statement": "int"}},
                    {"id": current_len + 2, "instrument": 'repeat', "action": 'endrepeat',
                     "args": {}, "return": '', "uuid": uid},
                ],
            "pause":
                [
                    {"id": current_len + 1, "instrument": 'pause', "action": "pause",
                     "args": {"statement": 1 if statement == '' else statement}, "return": '', "uuid": uid,
                     "arg_types": {"statement": "str"}}
                ],
            "comment":
                [
                    {"id": current_len + 1, "instrument": 'comment', "action": "comment",
                     "args": {"statement": statement}, "return": '', "uuid": uid,
                     "arg_types": {"statement": "str"}}
                ],
        }
        action_list = logic_dict[logic_type]
        self.currently_editing_script.extend(action_list)
        self._insert_action(insert_position, current_len, len(action_list))
        self.update_time_stamp()

    def delete_action(self, id: int):
        """
        Delete the action by id (step number)
        """
        uid = next((action['uuid'] for action in self.currently_editing_script if action['id'] == int(id)), None)
        id_to_be_removed = [action['id'] for action in self.currently_editing_script if action['uuid'] == uid]
        order = self.currently_editing_order
        script = self.currently_editing_script
        self.currently_editing_order = [i for i in order if int(i) not in id_to_be_removed]
        self.currently_editing_script = [action for action in script if action['id'] not in id_to_be_removed]
        self.sort_actions()
        self.update_time_stamp()

    def duplicate_action(self, id: int):
        """
        duplicate action by id (step number), available only for non logic actions
        """
        action_to_duplicate = next((action for action in self.currently_editing_script if action['id'] == int(id)),
                                   None)
        insert_id = action_to_duplicate.get("id")
        self.add_action(action_to_duplicate)
        # print(self.currently_editing_script)
        if action_to_duplicate is not None:
            # Update IDs for all subsequent actions
            for action in self.currently_editing_script:
                if action['id'] > insert_id:
                    action['id'] += 1
            self.currently_editing_script[-1]['id'] = insert_id + 1
            # Sort actions if necessary and update the time stamp
            self.sort_actions()
            self.update_time_stamp()
        else:
            raise ValueError("Action not found: Unable to duplicate the action with ID", id)

    def config(self, stype):
        """
        take the global script_dict
        :return: list of variable that require input
        """
        configure = []
        variables = self.get_variables()
        # print(variables)
        config_type_dict = {}
        for action in self.script_dict[stype]:
            args = action['args']
            if args is not None:
                if type(args) is not dict:
                    if type(args) is str and args.startswith("#"):
                        key = args[1:]
                        if key not in (*variables, *configure):
                            configure.append(key)
                            config_type_dict[key] = action['arg_types']

                else:
                    if action['instrument'] == "math_variable":
                        # assume any kind of math variable will be evaluated to float even if it might be an int
                        pattern = r"#([A-Za-z_][A-Za-z0-9_]*)"
                        vars_found = re.findall(pattern, args['statement'])
                        for key in vars_found:
                            if key not in (*variables, *configure):
                                configure.append(key)
                                config_type_dict[key] = action['arg_types']['statement']
                    else:
                        for arg in args:
                            if type(args[arg]) is str and args[arg].startswith("#"):
                                key = args[arg][1:]
                                if key not in (*variables, *configure):
                                    configure.append(key)
                                    if arg in action['arg_types']:
                                        if action['arg_types'][arg] == '':
                                            config_type_dict[key] = "any"
                                        else:
                                            config_type_dict[key] = action['arg_types'][arg]
                                    else:
                                        config_type_dict[key] = "any"
        # todo
        return configure, config_type_dict

    def config_return(self):
        """
        take the global script_dict
        :return: list of variable that require input
        """

        return_list = set([action['return'] for action in self.script_dict['script'] if not action['return'] == ''])
        output_str = "{"
        for i in return_list:
            output_str += "'" + i + "':" + i + ","
        output_str += "}"
        return output_str, return_list

    def finalize(self):
        """finalize script, disable editing"""
        self.status = "finalized"
        self.update_time_stamp()

    def save_as(self, name):
        """resave script, enable editing"""
        self.name = name
        self.status = "editing"
        self.update_time_stamp()

    def indent(self, unit=0):
        """helper: create _ unit of indent in code string"""
        string = "\n"
        for _ in range(unit):
            string += "\t"
        return string

    def convert_to_lines(self, exec_str_collection: dict):
        """
        Parse a dictionary of script functions and extract function body lines.

        :param exec_str_collection: Dictionary containing script types and corresponding function strings.
        :return: A dict containing script types as keys and lists of function body lines as values.
        """
        line_collection = {}

        for stype, func_str in exec_str_collection.items():
            if func_str:
                module = ast.parse(func_str)

                # Find the first function (regular or async)
                func_def = next(
                    node for node in module.body
                    if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
                )

                # Extract function body as source lines, skipping 'return' nodes
                line_collection[stype] = [
                    ast_unparse(node) for node in func_def.body if not isinstance(node, ast.Return)
                ]
        return line_collection

    def render_script_lines(self, script_dict):
        """
        Convert the script_dict structure into a dict of displayable Python-like lines,
        keeping ID consistency for highlighting.
        """

        def render_args(args):
            if not args:
                return ""
            return ", ".join(f"{k}={v}" for k, v in args.items())

        def parse_block(block):
            lines = []
            indent = 0
            stack = []

            for action in block:
                act = action["action"]
                _id = action["id"]
                instrument = action["instrument"]

                # Handle control structures
                if act == "if":
                    stmt = action["args"].get("statement", "")
                    lines.append("    " * indent + f"if {stmt}:")
                    indent += 1
                    stack.append("if")

                elif act == "else":
                    indent -= 1
                    lines.append("    " * indent + f"else:")
                    indent += 1

                elif act in ("endif", "endwhile"):
                    if stack:
                        stack.pop()
                    indent = max(indent - 1, 0)
                    lines.append("    " * indent + f"# {act}")

                elif act == "while":
                    stmt = action["args"].get("statement", "")
                    lines.append("    " * indent + f"while {stmt}:")
                    indent += 1
                    stack.append("while")

                elif act == "wait":
                    stmt = action["args"].get("statement", "")
                    if isinstance(stmt, str) and stmt.startswith("#"):
                        stmt = stmt[1:]
                    lines.append("    " * indent + f"time.sleep({stmt})")

                elif instrument == "variable":
                    stmt = action["args"].get("statement", "")
                    if isinstance(stmt, str) and stmt.startswith("#"):
                        stmt = stmt[1:]
                    lines.append("    " * indent + f"{act} = {stmt}")

                elif instrument == "input":
                    stmt = action["args"].get("statement", "")
                    var_name = action["args"].get("variable", "")
                    lines.append("    " * indent + f"{var_name} = input({repr(stmt)})")

                elif instrument == "math_variable":
                    stmt = action["args"].get("statement", "")
                    lines.append("    " * indent + f"{act} = {stmt}")


                else:
                    # Regular function call
                    instr = action["instrument"]
                    args = render_args(action.get("args", {}))
                    ret = action.get("return")
                    line = "    " * indent
                    if ret:
                        line += f"{ret} = {instr}.{act}({args})"
                    else:
                        line += f"{instr}.{act}({args})"
                    lines.append(line)

            # Ensure empty control blocks get "pass"
            final_lines = []
            for i, line in enumerate(lines):
                final_lines.append(line)
                # if line.strip().startswith("else") and (
                #         i == len(lines) - 1 or lines[i + 1].startswith("#") or "endif" in lines[i + 1]
                # ):
                    # final_lines.append("    " * (indent) + "pass")

            return final_lines

        return {
            "prep": parse_block(script_dict.get("prep", [])),
            "script": parse_block(script_dict.get("script", [])),
            "cleanup": parse_block(script_dict.get("cleanup", []))
        }

    def render_nested_script_lines(self, script_dict, snapshot=None):
        """
        Convert script_dict into a nested structure for collapsible UI.
        Returns check dict: { 'script': [ {type: 'line', code: '...', id: 'script-0'}, ... ] }
        """
        def render_args(args):
            if not args:
                return ""
            return ", ".join(f"{k}={v}" for k, v in args.items())

        def parse_block(block, parent_id_prefix, indent=0):
            nodes = []
            
            for i, action in enumerate(block):
                act = action["action"]
                instrument = action["instrument"]
                current_id = f"{parent_id_prefix}-{i}"
                
                # Handle nested workflows
                if instrument == 'workflows':
                    args = render_args(action.get("args", {}))
                    ret = action.get("return")
                    line_code = "    " * indent
                    if ret:
                        line_code += f"{ret} = "
                    line_code += f"{act}({args})"
                    
                    category = "workflow" # collapsible
                    
                    # Recursively parse children
                    # Prioritize embedded workflow steps
                    children_steps = action.get("workflow", [])
                    if not children_steps:
                        # Fallback to DB lookup if needed, though likely compilation ensures it's there
                        # but safe to check
                        wf_script = Script.query.filter_by(name=act).first()
                        if wf_script:
                            children_steps = wf_script.script_dict.get('script', [])

                    children_nodes = parse_block(children_steps, current_id, indent + 1)
                    
                    nodes.append({
                        "type": "workflow",
                        "code": line_code,
                        "id": current_id,
                        "children": children_nodes
                    })
                    
                # Handle control structures (simplified for now, treated as lines or blocks if we want collapse)
                # For now, let's treat control structures as flat or just lines, 
                # unless we want to collapse 'if' blocks too. 
                # User specifically asked for workflow collapse. Let's stick to that for now to avoid complexity.
                else: 
                     # Re-use the existing logic to generate the line string
                    line_code = ""
                    # ... [Logic to generate line string similar to render_script_lines] ...
                    # To avoid code duplication, I'll copy the logic briefly or helper
                    
                    if act == "if":
                        stmt = action["args"].get("statement", "")
                        line_code = "    " * indent + f"if {stmt}:"
                    elif act == "else":
                        # else is tricky because it belongs to previous if, but structurally it's a sibling in the list? 
                        # actually in the JSON list, else is likely a sibling.
                         line_code = "    " * (indent -1) + f"else:" if indent > 0 else "else:"
                    elif act in ("endif", "endwhile"):
                         line_code = "    " * (indent -1) + f"# {act}" if indent > 0 else f"# {act}"
                    elif act == "while":
                        stmt = action["args"].get("statement", "")
                        line_code = "    " * indent + f"while {stmt}:"
                    elif act == "wait":
                        stmt = action["args"].get("statement", "")
                        if isinstance(stmt, str) and stmt.startswith("#"):
                            stmt = stmt[1:]
                        line_code = "    " * indent + f"time.sleep({stmt})"
                    elif instrument == "variable":
                        stmt = action["args"].get("statement", "")
                        if isinstance(stmt, str) and stmt.startswith("#"):
                             stmt = stmt[1:]
                        line_code = "    " * indent + f"{act} = {stmt}"
                    elif instrument == "math_variable":
                        stmt = action["args"].get("statement", "")
                        line_code = "    " * indent + f"{act} = {stmt}"
                    elif instrument == "comment":
                        stmt = action["args"].get("statement", "")
                        if isinstance(stmt, str) and stmt.startswith("#"):
                             stmt = stmt[1:]
                             line_code = "    " * indent + f"print({stmt})"
                        else:
                             # Use repr to handle quotes safely
                             line_code = "    " * indent + f"print({repr(stmt)})"
                    elif instrument == "input":
                        stmt = action["args"].get("statement", "")
                        var_name = action["args"].get("variable", "var")
                        line_code = "    " * indent + f"{var_name} = input({repr(stmt)})"
                    else:
                        args_dict = action.get("args", {})
                        args = render_args(args_dict)
                        ret = action.get("return")
                        line_code = "    " * indent
                        
                        # Check metadata for properties
                        is_property_setter = False
                        is_property_getter = False
                        property_name = None
                        
                        if snapshot and instrument in snapshot:
                            inst_data = snapshot[instrument]
                            if act.endswith("_(setter)"):
                                prop_candidate = act[:-9]
                                if prop_candidate in inst_data and inst_data[prop_candidate].get('is_property'):
                                    is_property_setter = True
                                    property_name = prop_candidate
                            elif act in inst_data and inst_data[act].get('is_property'):
                                is_property_getter = True
                                property_name = act
                        
                        if is_property_setter:
                             arg_val = args_dict.get('value')
                             if isinstance(arg_val, str) and not arg_val.startswith("#"):
                                 arg_val = f"'{arg_val}'"
                             elif isinstance(arg_val, str) and arg_val.startswith("#"):
                                 arg_val = arg_val[1:]
                             line_code += f"{instrument}.{property_name} = {arg_val}"
                        elif is_property_getter:
                             if ret:
                                 line_code += f"{ret} = {instrument}.{property_name}"
                             else:
                                 line_code += f"{instrument}.{property_name}"
                        else:
                            if ret:
                                line_code += f"{ret} = {instrument}.{act}({args})"
                            else:
                                line_code += f"{instrument}.{act}({args})"
                    
                    nodes.append({
                        "type": "line",
                        "code": line_code,
                        "id": current_id
                    })
                    
            return nodes

        return {
            "prep": parse_block(script_dict.get("prep", []), "prep"),
            "script": parse_block(script_dict.get("script", []), "script"),
            "cleanup": parse_block(script_dict.get("cleanup", []), "cleanup")
        }

    def compile(self, script_path=None, batch=False, mode="sample", snapshot=None):
        """
        Compile the current script to a Python file.
        :return: String to write to a Python file.
        """
        self.needs_call_human = False
        self.blocks_included = False

        self.sort_actions()
        run_name = self.name if self.name else "untitled"
        run_name = self.validate_function_name(run_name)
        exec_str_collection = {}

        for i in self.stypes:
            if self.script_dict[i]:
                is_async = any(a.get("coroutine", False) for a in self.script_dict[i])
                func_str = self._generate_function_header(run_name, i, is_async, batch) + self._generate_function_body(i, batch, mode, snapshot)
                exec_str_collection[i] = func_str
        if script_path:
            self._write_to_file(script_path, run_name, exec_str_collection)

        return exec_str_collection



    @staticmethod
    def validate_function_name(name):
        """Replace invalid characters with underscores"""
        name = re.sub(r'\W|^(?=\d)', '_', name)
        # Check if it's a Python keyword and adjust if necessary
        if keyword.iskeyword(name):
            name += '_'
        return name

    def _generate_function_header(self, run_name, stype, is_async, batch=False):
        """
        Generate the function header.
        """
        configure, config_type = self.config(stype)
        configure = [param + f":{param_type}" if not param_type == "any" else param for param, param_type in
                     config_type.items()]

        script_type = f"_{stype}" if stype != "script" else ""
        async_str = "async " if is_async else ""
        function_header = f"{async_str}def {run_name}{script_type}("

        if stype == "script":
            if batch:
                function_header += "param_list" if configure else "n: int"
            else:
                function_header += ", ".join(configure)
        function_header += "):"

        if stype == "script" and batch:
            function_header += self.indent(1) + f'"""Batch mode is experimental and may have bugs."""'
        return function_header

    def _generate_function_body(self, stype, batch=False, mode="sample", snapshot=None):
        """
        Generate the function body for each type in stypes.
        """
        body = ''
        indent_unit = 1
        if batch and stype == "script":
            return_str, return_list = self.config_return()
            configure, config_type = self.config(stype)
            if not configure:
                body += self.indent(indent_unit) + "param_list = [{} for _ in range(n)]"
            for index, action in enumerate(self.script_dict[stype]):
                text, indent_unit = self._process_action(indent_unit, action, index, stype, batch, snapshot=snapshot)
                body += text
            if return_list:
                # body += self.indent(indent_unit) + f"result_list.append({return_str})"
                body += self.indent(indent_unit) + "return param_list"

        else:
            for index, action in enumerate(self.script_dict[stype]):
                text, indent_unit = self._process_action(indent_unit, action, index, stype, snapshot=snapshot)
                body += text
            return_str, return_list = self.config_return()
            if return_list and stype == "script":
                body += self.indent(indent_unit) + f"return {return_str}"
        return body

    def _process_action(self, indent_unit, action, index, stype, batch=False, mode="sample", snapshot=None):
        """
        Process each action within the script dictionary.
        """
        configure, config_type = self.config(stype)

        instrument = action['instrument']
        statement = action['args'].get('statement')
        args = self._process_args(action['args'])

        save_data = action['return']
        action_name = action['action']
        batch_action = action.get("batch_action", False)

        next_action = self._get_next_action(stype, index)
        # print(args)
        if instrument == 'if':
            return self._process_if(indent_unit, action_name, statement, next_action)
        elif instrument == 'while':
            return self._process_while(indent_unit, action_name, statement, next_action)
        elif instrument == 'variable':
            if batch:
                if isinstance(statement, str) and statement.startswith("#"):
                    return '', indent_unit
                return self.indent(indent_unit) + "for param in param_list:" + self.indent(
                    indent_unit + 1) + f"param['{action_name}'] = {statement}", indent_unit
            else:
                if isinstance(statement, str) and statement.startswith("#"):
                    statement = statement[1:]
                return self.indent(indent_unit) + f"{action_name} = {statement}", indent_unit
        elif instrument == 'input':
            var_name = args.get('variable')
            var_type = args.get('variable_type', 'str')
            prompt = statement
            
            # Construct input call
            if isinstance(prompt, str) and prompt.startswith("#"):
                prompt_expr = prompt[1:]
            else:
                prompt_expr = repr(prompt) if prompt else "''"
                
            input_call = f"input({prompt_expr})"
            
            # Apply type casting
            if var_type == 'int':
                expr = f"int({input_call})"
            elif var_type == 'float':
                expr = f"float({input_call})"
            elif var_type == 'bool':
            # Simplified bool conversion (non-empty string is True) or specific 'true' check?
            # Standard bool("False") is True. 
            # Let's simple use bool() for now or specific parsing if needed. 
            # Python input() returns string. 
                expr = f"bool({input_call})"
            else:
                expr = input_call
                
            if batch:
                return self.indent(indent_unit) + "for param in param_list:" + \
                       self.indent(indent_unit + 1) + f"param['{var_name}'] = {expr}", indent_unit
            else:
                return self.indent(indent_unit) + f"{var_name} = {expr}", indent_unit

        elif instrument == 'wait':
            if isinstance(statement, str) and statement.startswith("#"):
                statement = statement[1:]
            if batch:
                return f"{self.indent(indent_unit)}for param in param_list:" + f"{self.indent(indent_unit+1)}time.sleep({statement})", indent_unit
            return f"{self.indent(indent_unit)}time.sleep({statement})", indent_unit
        elif instrument == 'repeat':
            return self._process_repeat(indent_unit, action_name, statement, next_action)
        elif instrument == 'comment':
            if isinstance(statement, str) and statement.startswith("#"):
                statement = statement[1:]
                out_stmt = f"print({statement})"
            else:
                # Use repr to safely escape string with quotes
                out_stmt = f"print({repr(statement)})"

            if batch:
                return f"{self.indent(indent_unit)}for param in param_list:" + f"{self.indent(indent_unit + 1)}{out_stmt}", indent_unit
            return f"{self.indent(indent_unit)}{out_stmt}", indent_unit
        elif instrument == 'pause':
            self.needs_call_human = True
            if isinstance(statement, str) and statement.startswith("#"):
                statement = f"str({statement[1:]})"
            else:
                # cases like "text with a backslash \"
                statement = statement.encode('unicode_escape').decode()
            if batch:
                return f"{self.indent(indent_unit)}for param in param_list:" + f"{self.indent(indent_unit+1)}pause('''{statement}''')", indent_unit
            return f"{self.indent(indent_unit)}pause('''{statement}''')", indent_unit
        elif instrument == "math_variable":
            math_expression = self._process_math(statement)
            if batch:
                return f"{self.indent(indent_unit)}for param in param_list:" + f"{self.indent(indent_unit + 1)}param['{action_name}'] = {math_expression}", indent_unit
            else:
                return f"{self.indent(indent_unit)}{action_name} = {math_expression}", indent_unit
        # todo
        # elif instrument == 'registered_workflows':
        #     return inspect.getsource(my_function)
        else:
            is_async = action.get("coroutine", False)
            dynamic_arg = len(self.get_variables()) > 0
            workflow_steps = action.get("workflow", None)
        return self._process_instrument_action(indent_unit, instrument, action_name, args, save_data, is_async, dynamic_arg, batch, batch_action, workflow_steps, snapshot=snapshot)

    def _process_args(self, args):
        """
        Process arguments, handling any specific formatting needs.
        """
        if isinstance(args, str) and args.startswith("#"):
            return args[1:]
        return args

    def _process_if(self, indent_unit, action, args, next_action):
        """
        Process 'if' and 'else' actions.
        """
        exec_string = ""
        if action == 'if':
            if isinstance(args, str) and args.startswith("#"):
                args = args[1:]
            exec_string += self.indent(indent_unit) + f"if {args}:"
            indent_unit += 1
            if next_action and next_action['instrument'] == 'if' and next_action['action'] == 'else':
                exec_string += self.indent(indent_unit) + "pass"
            # else:

        elif action == 'else':
            indent_unit -= 1
            exec_string += self.indent(indent_unit) + "else:"
            indent_unit += 1
            if next_action and next_action['instrument'] == 'if' and next_action['action'] == 'endif':
                exec_string += self.indent(indent_unit) + "pass"
        else:
            indent_unit -= 1
        return exec_string, indent_unit

    def _process_while(self, indent_unit, action, args, next_action):
        """
        Process 'while' and 'endwhile' actions.
        """
        exec_string = ""
        if action == 'while':
            if isinstance(args, str) and args.startswith("#"):
                args = args[1:]
            exec_string += self.indent(indent_unit) + f"while {args}:"
            indent_unit += 1
            if next_action and next_action['instrument'] == 'while':
                exec_string += self.indent(indent_unit) + "pass"
        elif action == 'endwhile':
            indent_unit -= 1
        return exec_string, indent_unit

    def _process_repeat(self, indent_unit, action, args, next_action):
        """
        Process 'while' and 'endwhile' actions.
        """
        exec_string = ""
        if isinstance(args, str) and args.startswith("#"):
            args = args[1:]
        if action == 'repeat':
            exec_string += self.indent(indent_unit) + f"for _ in range({args}):"
            indent_unit += 1
            if next_action and next_action['instrument'] == 'repeat':
                exec_string += self.indent(indent_unit) + "pass"
        elif action == 'endrepeat':
            indent_unit -= 1
        return exec_string, indent_unit

    def _process_math(self, expr: str) -> str:
        """
        Return the math expression as a string but remove '#' prefix from variable names
        """
        # remove leading "#" for variable tokens
        cleaned = re.sub(r"#([A-Za-z_]\w*)", r"\1", expr)
        return cleaned

    def _process_instrument_action(self, indent_unit, instrument, action, args, save_data, is_async=False, dynamic_arg=False,
                                   batch=False, batch_action=False, workflow_steps=None, snapshot=None):
        """
        Process actions related to instruments.
        """
        async_str = "await " if is_async else ""

        if instrument == 'workflows':
            # This is a call to another registered workflow
            # Check for embedded steps first, otherwise fetch
            if workflow_steps is not None:
                # Use embedded steps (already a list of dicts)
                script_actions = workflow_steps
                # Create a temporary Script object just to use its context/methods if needed? 
                # Actually we can use 'self' but we need to ensure the methods called inside are generic.
                # _process_action uses self.script_dict etc. 
                # Wait, if we recurse, we need to be careful.
                # Let's inspect logic below.
                
                # We need a context for the inner workflow?
                # Using 'self' is fine for _process_action as long as it doesn't depend on self.script_dict for the *current* action being processed (it's passed in).
                # But _process_action requires 'stype' which is usually 'script'.
                pass # placeholder for logic logic
            else:
                 workflow_script = Script.query.get(action)
                 if workflow_script:
                     script_actions = workflow_script.script_dict.get('script', [])
                 else:
                     script_actions = []

            if script_actions:
                output_code = self.indent(indent_unit) + f"# Begin Workflow: {action}"
                
                # Inject argument assignments
                if args and isinstance(args, dict):
                     for key, value in args.items():
                         if isinstance(value, str) and value.startswith("#"):
                             # Passing a variable: inner_var = outer_var
                             assignment = f"{key} = {value[1:]}"
                         elif isinstance(value, str):
                             # String literal
                             assignment = f"{key} = '{value}'"
                         else:
                             # Other literals (int, float, bool)
                             assignment = f"{key} = {value}"
                         output_code += self.indent(indent_unit) + assignment

                expanded_body = ""
                # Use 'self' to process the inner actions. 
                # Prerequisite: _process_action is stateless regarding the script content list.
                # It seems so.
                for i, inner_action in enumerate(script_actions):
                     text, _ = self._process_action(indent_unit, inner_action, i, 'script', batch, mode="sample") 
                     expanded_body += text
                
                output_code += expanded_body
                output_code += self.indent(indent_unit) + f"# End Workflow: {action}"
                return output_code, indent_unit

        function_call = f"{instrument}.{action}"
        if instrument.startswith("blocks"):
            self.blocks_included = True
            function_call = action

        # Check if this is a property getter/setter
        is_property_setter = False
        is_property_getter = False
        property_name = None
        
        if snapshot and instrument in snapshot:
            inst_data = snapshot[instrument]
            if action.endswith("_(setter)"):
                prop_candidate = action[:-9]
                if prop_candidate in inst_data and inst_data[prop_candidate].get('is_property'):
                     is_property_setter = True
                     property_name = prop_candidate
            elif action in inst_data and inst_data[action].get('is_property'):
                is_property_getter = True
                property_name = action

        if is_property_setter:
            # Generate assignment: inst.prop = value
            arg_str = "None"
            if isinstance(args, dict):
                 val = args.get('value')
                 if val is not None:
                      # If it is a string starting with # or text, handle quoting
                      if isinstance(val, str):
                           if val.startswith("#"):
                                arg_str = val[1:]
                           else:
                                arg_str = f"'{val}'"
                      else:
                           arg_str = str(val)
            single_line = f"{async_str}{instrument}.{property_name} = {arg_str}"

        elif is_property_getter:
            # Generate access: inst.prop (no parentheses)
            single_line = f"{async_str}{instrument}.{property_name}"

        elif isinstance(args, dict) and args != {}:
            args_str = self._process_dict_args(args)
            single_line = f"{async_str}{function_call}(**{args_str})"
        elif isinstance(args, str):
            single_line = f"{function_call} = {args}"
        else:
            single_line = f"{async_str}{function_call}()"


        save_data_str = save_data + " = " if save_data else ''

        if batch and not batch_action:
            arg_list = [args[arg][1:] for arg in args if isinstance(args[arg], str) and args[arg].startswith("#")]
            param_str = [f"param['{arg_list}']" for arg_list in arg_list if arg_list]
            args_str = self.indent(indent_unit + 1) +  ", ".join(arg_list) + " = " + ", ".join(param_str) if arg_list else ""
            if dynamic_arg:
                for_string = self.indent(indent_unit) + "for param in param_list:" + args_str
            else:
                for_string = self.indent(indent_unit) + "for i in range(n):"
            output_code = for_string + self.indent(indent_unit + 1) + save_data_str + single_line
            if save_data:
                output_code = output_code + self.indent(indent_unit + 1) + f"param['{save_data}'] = {save_data}"
        else:
            output_code = self.indent(indent_unit) + save_data_str + single_line

        return output_code, indent_unit

    def _process_dict_args(self, args):
        """
        Process dictionary arguments, handling special cases like variables.
        """
        args_str = args.__str__()
        for arg in args:
            if isinstance(args[arg], str) and args[arg].startswith("#"):
                args_str = args_str.replace(f"'#{args[arg][1:]}'", args[arg][1:])
            elif isinstance(args[arg], dict):
                # print(args[arg])
                if not args[arg]:
                    continue
                # Extract the variable name (first key in the dict)
                value = next(iter(args[arg]))
                var_type = args[arg].get(value)

                # Only process if it's a function_output variable reference
                if var_type == "function_output":
                    variables = self.get_variables()
                    if value not in variables:
                        raise ValueError(f"Variable ({value}) is not defined.")
                    # Replace the dict string representation with just the variable name
                    args_str = args_str.replace(f"{args[arg]}", value)

        # elif self._is_variable(arg):
            #     print("is variable")
            #     args_str = args_str.replace(f"'{args[arg]}'", args[arg])
        return args_str

    def _get_next_action(self, stype, index):
        """
        Get the next action in the sequence if it exists.
        """
        if index < (len(self.script_dict[stype]) - 1):
            return self.script_dict[stype][index + 1]
        return None

    def _is_variable(self, arg):
        """
        Check if the argument is of type 'variable'.
        """
        return arg in self.script_dict and self.script_dict[arg].get("arg_types") in ("variable", 'math_variable')

    def _write_to_file(self, script_path, run_name, exec_string, call_human=False):
        """
        Write the compiled script to a file.
        """
        with open(script_path + run_name + ".py", "w") as s:
            if self.deck:
                s.write(f"import {self.deck} as deck")
            else:
                s.write("deck = None")
            s.write("\nimport time")
            if self.blocks_included:
                s.write(f"\n{self._create_block_import()}")
            if self.needs_call_human:
                s.write("""\n\ndef pause(reason="Manual intervention required"):\n\tprint(f"\\nHUMAN INTERVENTION REQUIRED: {reason}")\n\tinput("Press Enter to continue...\\n")""")

            for i in exec_string.values():
                s.write(f"\n\n\n{i}")

    def add_workflow(self, workflow_name, insert_position=None, **kwargs):
        current_len = len(self.currently_editing_script)
        uid = uuid.uuid4().fields[-1]
        action = {"id": current_len + 1, "instrument": "deck_name", "action": workflow_name,
                  "args": kwargs, "return": '', "uuid": uid,
                  "arg_types": {}}
        self.currently_editing_script.append(action)
        self._insert_action(insert_position, current_len)
        self.update_time_stamp()

    def group_actions(self, ids: list, name: str, author: str):
        """
        Group selected actions into a new registered workflow script.

        :param ids: List of action IDs (step numbers) to group.
        :param name: Name of the new workflow.
        :param author: Author of the new workflow.
        """
        # 1. Validate inputs
        if not ids:
            return
        
        # Sort ids to ensure we process them in order
        ids = sorted([int(id) for id in ids])
        
        # 2. Extract actions to be grouped
        actions_to_group = []
        for action in self.currently_editing_script:
            if action['id'] in ids:
                actions_to_group.append(action)
        
        # Sort actions by their ID to maintain order
        actions_to_group.sort(key=lambda x: x['id'])

        if not actions_to_group:
            return

        # 3. Create new Script for the grouped workflow
        # We assume the new script should have the same deck as the current one
        new_script_dict = {"prep": [], "cleanup": [], "script": []}
        
        # Deep copy actions to avoid reference issues, and reset their IDs
        import copy
        new_actions = copy.deepcopy(actions_to_group)
        for i, action in enumerate(new_actions):
            action['id'] = i + 1
            # Keep UUIDs or generate new ones? 
            # Generating new ones is safer to avoid duplicates across scripts if that matters
            action['uuid'] = uuid.uuid4().fields[-1] 
        
        new_script_dict["script"] = new_actions
        
        new_script = Script(
            name=name,
            deck=self.deck,
            status="finalized", # Or "editing"? Maybe finalized so it's immediately usable?
            script_dict=new_script_dict,
            author=author,
            # registered=True,
            description=f"Grouped from {len(ids)} steps"
        )
        
        # Save the new script to DB - THIS NEEDS DB SESSION ACCESS usually.
        # But Script object methods here seem to only modify self state usually.
        # We need to return this new script object so the caller can save it to DB.
        
        # 4. Remove original actions from current script
        # We need to insert the new workflow action at the position of the *first* removed action
        first_id = ids[0]
        insert_index = -1
        
        # Find index in currently_editing_script
        for i, action in enumerate(self.currently_editing_script):
            if action['id'] == first_id:
                insert_index = i
                break
        
        # Filter out grouped actions
        self.currently_editing_script = [a for a in self.currently_editing_script if a['id'] not in ids]
        
        # 5. Insert new workflow call
        # We need to construct arguments for the workflow if any internal variables are used?
        # For now, let's assume no arguments or let user configure later.
        # But wait, if the grouped steps used variables, those variables might need to be passed in.
        # Complex variable analysis is hard. For v1, let's assume empty args.
        
        # uuid for the new action
        uid = uuid.uuid4().fields[-1]
        
        workflow_action = {
            "id": 0, # Will be fixed by sort/insert logic or we set it now? self.add_workflow handles appending.
            "instrument": "deck_name", # This seems to be the convention for workflows?
            "action": name,
            "args": {},
            "return": "",
            "uuid": uid,
            "arg_types": {}
        }
        
        # We can't use self.add_workflow directly easily because we want to insert at specific index 
        # NOT just append and then invalidly insert. 
        # Actually `add_workflow` logic: append then `_insert_action`.
        # `_insert_action` logic: manipulation of `currently_editing_order`.
        
        # Let's direct modify `currently_editing_script` and `currently_editing_order`.
        
        self.currently_editing_script.insert(insert_index, workflow_action)
        
        # 6. Fix IDs
        # We simply re-assign IDs based on position
        for i, action in enumerate(self.currently_editing_script):
            action['id'] = i + 1
            
        # 7. Update Order
        # Reset order to match current script list
        self.currently_editing_order = [str(a['id']) for a in self.currently_editing_script]
        
        self.update_time_stamp()
        
        return new_script

    def _create_block_import(self):
        imports = {}
        from ivoryos.utils.decorators import BUILDING_BLOCKS
        for category, methods in BUILDING_BLOCKS.items():
            for method_name, meta in methods.items():
                func = meta["func"]
                module = meta["path"]
                name = func.__name__
                imports.setdefault(module, set()).add(name)
        lines = []
        for module, funcs in imports.items():
            lines.append(f"from {module} import {', '.join(sorted(funcs))}")
        return "\n".join(lines)

class WorkflowRun(db.Model):
    """Represents the entire experiment"""
    __tablename__ = 'workflow_runs'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(128), nullable=False)
    platform = db.Column(db.String(128), nullable=False)
    start_time = db.Column(db.DateTime, default=datetime.now())
    end_time = db.Column(db.DateTime)
    data_path = db.Column(db.String(256))
    repeat_mode = db.Column(db.String(64), default="none")  # static_repeat, sweep, optimizer

    # A run contains multiple iterations
    phases = db.relationship(
        'WorkflowPhase',
        backref='workflow_runs', # Clearer back-reference name
        cascade='all, delete-orphan',
        lazy='dynamic' # Good for handling many iterations
    )
    def as_dict(self):
        dict = self.__dict__
        dict.pop('_sa_instance_state', None)
        return dict

class WorkflowPhase(db.Model):
    """Represents a single function call within a WorkflowRun."""
    __tablename__ = 'workflow_phases'

    id = db.Column(db.Integer, primary_key=True)
    # Foreign key to link this iteration to its parent run
    run_id = db.Column(db.Integer, db.ForeignKey('workflow_runs.id', ondelete='CASCADE'), nullable=False)

    # NEW: Store iteration-specific parameters here
    name = db.Column(db.String(64), nullable=False)  # 'prep', 'main', 'cleanup'
    repeat_index = db.Column(db.Integer, default=0)

    parameters = db.Column(JSONType)  # Use db.JSON for general support
    outputs = db.Column(JSONType)
    start_time = db.Column(db.DateTime, default=datetime.now)
    end_time = db.Column(db.DateTime)

    # An iteration contains multiple steps
    steps = db.relationship(
        'WorkflowStep',
        backref='workflow_phases',  # Clearer back-reference name
        cascade='all, delete-orphan'
    )

    def as_dict(self):
        dict = self.__dict__.copy()
        dict.pop('_sa_instance_state', None)
        return dict

class WorkflowStep(db.Model):
    __tablename__ = 'workflow_steps'

    id = db.Column(db.Integer, primary_key=True)
    # workflow_id = db.Column(db.Integer, db.ForeignKey('workflow_runs.id', ondelete='CASCADE'), nullable=True)
    phase_id = db.Column(db.Integer, db.ForeignKey('workflow_phases.id', ondelete='CASCADE'), nullable=True)

    # phase = db.Column(db.String(64), nullable=False)  # 'prep', 'main', 'cleanup'
    # repeat_index = db.Column(db.Integer, default=0)   # Only applies to 'main' phase
    step_index = db.Column(db.Integer, default=0)
    method_name = db.Column(db.String(128), nullable=False)
    start_time = db.Column(db.DateTime)
    end_time = db.Column(db.DateTime)
    run_error = db.Column(db.Boolean, default=False)
    output = db.Column(JSONType, default={})
    # Using as_dict method from ModelBase

    def as_dict(self):
        dict = self.__dict__.copy()
        dict.pop('_sa_instance_state', None)
        return dict


class SingleStep(db.Model):
    __tablename__ = 'single_steps'

    id = db.Column(db.Integer, primary_key=True)
    method_name = db.Column(db.String(128), nullable=False)
    kwargs = db.Column(JSONType, nullable=False)
    start_time = db.Column(db.DateTime)
    end_time = db.Column(db.DateTime)
    run_error = db.Column(db.String(128))
    output = db.Column(JSONType, nullable=True)

    def as_dict(self):
        dict = self.__dict__.copy()
        dict.pop('_sa_instance_state', None)
        return dict

if __name__ == "__main__":
    a = Script()

    print("")



import copy

from flask import Blueprint, redirect, flash, request, render_template, session, current_app, jsonify
from flask_login import login_required

from ivoryos.routes.control.control_file import control_file
from ivoryos.routes.control.control_new_device import control_temp
from ivoryos.routes.control.utils import post_session_by_instrument, get_session_by_instrument, find_instrument_by_name
from ivoryos.utils.global_config import GlobalConfig
from ivoryos.utils.form import create_form_from_module, create_form_from_pseudo
from ivoryos.utils.task_runner import TaskRunner

global_config = GlobalConfig()
runner = TaskRunner()

control = Blueprint('control', __name__, template_folder='templates')

control.register_blueprint(control_file)
control.register_blueprint(control_temp)



@control.route("/", strict_slashes=False, methods=["GET", "POST"])
@control.route("/<string:instrument>", strict_slashes=False, methods=["GET", "POST"])
@login_required
async def deck_controllers(instrument: str = None):
    """
    .. :quickref: Direct Control; device (instruments) and methods

    device home interface for listing all instruments and methods, selecting an instrument to run its methods

    .. http:get:: /instruments

        get all instruments for home page

    .. http:get:: /instruments/<string:instrument>

        get all methods of the given <instrument>

    .. http:post:: /instruments/<string:instrument>

        send POST request to run a method of the given <instrument>

    :param instrument: instrument name, if not provided, list all instruments
    :type instrument: str
    :status 200: render template with instruments and methods

    """
    instrument = instrument or request.args.get("instrument")
    forms = None
    if instrument:
        inst_object = find_instrument_by_name(instrument)
        if instrument.startswith("blocks"):
            forms = create_form_from_pseudo(pseudo=inst_object, autofill=False, design=False)
        elif instrument.startswith("deck"):
            forms = create_form_from_pseudo(pseudo=global_config.deck_snapshot[instrument], autofill=False, design=False)
        else:
            #TODO
            forms = create_form_from_module(sdl_module=inst_object, autofill=False, design=False)
        order = get_session_by_instrument('card_order', instrument)
        hidden_functions = get_session_by_instrument('hidden_functions', instrument)
        functions = list(forms.keys())
        for function in functions:
            if function not in hidden_functions and function not in order:
                order.append(function)
        post_session_by_instrument('card_order', instrument, order)
        forms = {name: forms[name] for name in order if name in forms}

    if request.method == "POST":
        if not forms:
            return jsonify({"success": False, "error": "Instrument not found"}), 404

        payload = request.get_json() if request.is_json else request.form.to_dict()
        method_name = payload.pop("hidden_name", None)
        form = forms.get(method_name)

        if not form:
            return jsonify({"success": False, "error": f"Method {method_name} not found"}), 404

        # Extract kwargs
        if request.is_json:
            kwargs = {k: v for k, v in payload.items() if k not in ["csrf_token", "hidden_wait"]}
        else:
            if not form.validate_on_submit():
                flash(f"Run Error! {form.errors}", "error")
                return render_template(
                    "controllers.html",
                    defined_variables=global_config.deck_snapshot.keys(),
                    block_variables=global_config.building_blocks.keys(),
                    temp_variables=global_config.defined_variables.keys(),
                    instrument=instrument,
                    forms=forms,
                    session=session
                )
            else:
                kwargs = {field.name: field.data for field in form if field.name not in ["csrf_token", "hidden_name"]}

        wait = str(payload.get("hidden_wait", "true")).lower() == "true"

        output = await runner.run_single_step(
            component=instrument, method=method_name, kwargs=kwargs, wait=wait,
            current_app=current_app._get_current_object()
        )

        if request.is_json:
            return jsonify(output)
        else:
            if output.get("success"):
                flash(f"Run Success! Output: {output.get('output', 'None')}")
            else:
                flash(f"Run Error! {output.get('output', 'Unknown error occurred.')}", "error")

    # GET request → render web form or return snapshot for API
    if request.is_json or request.accept_mimetypes.best_match(['application/json', 'text/html']) == 'application/json':
        # 1.3.2 fix snapshot copy, add building blocks to snapshots
        snapshot = copy.deepcopy(global_config.deck_snapshot)
        building_blocks = copy.deepcopy(global_config.building_blocks)
        snapshot.update(building_blocks)
        for instrument_key, instrument_data in snapshot.items():
            for function_key, function_data in instrument_data.items():
                function_data["signature"] = str(function_data["signature"])
        return jsonify(snapshot)

    return render_template(
        "controllers.html",
        defined_variables=global_config.deck_snapshot.keys(),
        block_variables=global_config.building_blocks.keys(),
        temp_variables=global_config.defined_variables.keys(),
        instrument=instrument,
        forms=forms,
        session=session
    )

@control.route('/<string:instrument>/actions/order', methods=['POST'])
def save_order(instrument: str):
    """
    .. :quickref: Control Customization; Save functions' order

    .. http:post:: instruments/<string:instrument>/actions/order

    save function drag and drop order for the given <instrument>

    """
    # Save the new order for the specified group to session
    data = request.json
    post_session_by_instrument('card_order', instrument, data['order'])
    return '', 204

@control.route('/<string:instrument>/actions/<string:function>', methods=["PATCH"])
def hide_function(instrument: str, function: str):
    """
    .. :quickref: Control Customization; Toggle function visibility

    .. http:patch:: /instruments/<instrument>/actions/<function>

    Toggle visibility for the given <instrument> and <function>

    """
    back = request.referrer
    data = request.get_json()
    hidden = data.get('hidden', True)
    functions = get_session_by_instrument("hidden_functions", instrument)
    order = get_session_by_instrument("card_order", instrument)
    if hidden and function not in functions:
        functions.append(function)
        if function in order:
            order.remove(function)
    elif not hidden and function in functions:
        functions.remove(function)
        if function not in order:
            order.append(function)
    post_session_by_instrument('hidden_functions', instrument, functions)
    post_session_by_instrument('card_order', instrument, order)
    return jsonify(success=True, message="Visibility updated")








### ivoryos/optimizers/base.py

from abc import ABC, abstractmethod



class OptimizerBase(ABC):
    def __init__(self, experiment_name:str, parameter_space: list, objective_config: dict, optimizer_config: dict,
                 parameter_constraints:list=None, datapath:str=None, additional_params:dict=None):
        """
        :param experiment_name: arbitrary name
        :param parameter_space: list of parameter names
        [
            {"name": "param_1", "type": "range", "bounds": [1.0, 2.0], "value_type": "float"},
            {"name": "param_2", "type": "choice", "bounds": ["a", "b", "c"], "value_type": "str"},
            {"name": "param_3", "type": "range", "bounds": [0 10], "value_type": "int"},
        ]
        :param objective_config: objective configuration
                [
            {"name": "obj_1", "minimize": True, "weight": 1},
            {"name": "obj_2", "minimize": False, "weight": 1}
        ]
        :param optimizer_config: optimizer configuration
        optimizer_config={
            "step_1": {"model": "Random", "num_samples": 10},
            "step_2": {"model": "BOTorch"}
        }
        """
        self.experiment_name = experiment_name
        self.parameter_space = parameter_space
        self.objective_config = objective_config
        self.optimizer_config = optimizer_config
        self.parameter_constraints = parameter_constraints
        self.additional_params = additional_params
        self.datapath = datapath

    @abstractmethod
    def suggest(self, n=1):
        pass

    @abstractmethod
    def observe(self, results: dict):
        """
        observe
        :param results: {"objective_name": "value"}
        """
        pass

    @abstractmethod
    def append_existing_data(self, existing_data, file_path: str = None):
        pass

    @abstractmethod
    def get_plots(self, plot_type):
        pass

    @staticmethod
    def _create_discrete_search_space(range_with_step=None, value_type ="float"):
        if range_with_step is None:
            range_with_step = []
        import numpy as np
        low, high, step = range_with_step
        values = np.arange(low, high + 1e-9 * step, step).tolist()
        if value_type == "float":
            values = [float(v) for v in values]
        if value_type == "int":
            values = [int(v) for v in values]
        return values

    @staticmethod
    def get_schema():
        """
        Returns a template for the optimizer configuration.
        """
        return {
            "parameter_types": ["range", "choice"],
            "multiple_objectives": True,
            # "objective_weights": True,
            "optimizer_config": {
                "step_1": {"model": [], "num_samples": 10},
                "step_2": {"model": []}
            },
            "additional_field": {}
        }






import os

from flask import Blueprint, redirect, url_for, request, render_template, current_app, jsonify, send_file
from flask_login import login_required

from ivoryos.utils.db_models import db, WorkflowRun, WorkflowStep, WorkflowPhase

data = Blueprint('data', __name__, template_folder='templates')



@data.route('/executions/records')
@login_required
def list_workflows():
    """
    .. :quickref: Workflow Execution Database; list all workflow execution records

    list all workflow execution records

    .. http:get:: /executions/records

    """
    query = WorkflowRun.query.order_by(WorkflowRun.id.desc())
    search_term = request.args.get("keyword", None)
    if search_term:
        query = query.filter(WorkflowRun.name.like(f'%{search_term}%'))
    page = request.args.get('page', default=1, type=int)
    per_page = 10

    workflows = query.paginate(page=page, per_page=per_page, error_out=False)
    if request.accept_mimetypes.best_match(['application/json', 'text/html']) == 'application/json':
        workflows = query.all()
        workflow_data = {w.id:{"workflow_name":w.name, "start_time":w.start_time} for w in workflows}
        return jsonify({
            "workflow_data": workflow_data,
        })
    else:
        return render_template('workflow_database.html', workflows=workflows)

@data.get("/executions/records/<int:workflow_id>")
def workflow_logs(workflow_id:int):
    """
    .. :quickref: Workflow Data Database; get workflow data, steps, and logs

    get workflow data logs by workflow id

    .. http:get:: /executions/records/<int:workflow_id>

    :param workflow_id: workflow id
    :type workflow_id: int
    """
    workflow = db.session.get(WorkflowRun, workflow_id)
    if not workflow:
        return jsonify({"error": "Workflow not found"}), 404

    # Query all phases for this run, ordered by start_time
    phases = WorkflowPhase.query.filter_by(run_id=workflow_id).order_by(WorkflowPhase.start_time).all()

    # Prepare grouped data for template (full objects)
    grouped = {
        "prep": [],
        "script": {},
        "cleanup": [],
    }

    # Prepare grouped data for JSON (dicts)
    grouped_json = {
        "prep": [],
        "script": {},
        "cleanup": [],
    }

    for phase in phases:
        phase_dict = phase.as_dict()

        # Steps sorted by step_index
        steps = sorted(phase.steps, key=lambda s: s.step_index)
        phase_steps_dicts = [s.as_dict() for s in steps]

        if phase.name == "prep":
            grouped["prep"].append(phase)
            grouped_json["prep"].append({
                **phase_dict,
                "steps": phase_steps_dicts
            })

        elif phase.name == "main":
            grouped["script"].setdefault(phase.repeat_index, []).append(phase)
            grouped_json["script"].setdefault(phase.repeat_index, []).append({
                **phase_dict,
                "steps": phase_steps_dicts
            })

        elif phase.name == "cleanup":
            grouped["cleanup"].append(phase)
            grouped_json["cleanup"].append({
                **phase_dict,
                "steps": phase_steps_dicts
            })

    if request.accept_mimetypes.best_match(['application/json', 'text/html']) == 'application/json':
        return jsonify({
            "workflow_info": workflow.as_dict(),
            "phases": grouped_json,
            "csv_file_name": f"{workflow.data_path}"
        })
    else:
        return render_template("workflow_view.html", workflow=workflow, grouped=grouped)


@data.get("/executions/data/<int:workflow_id>")
def workflow_phase_data(workflow_id: int):
    """
    .. :quickref: Workflow Data Database; get workflow data for plotting

    get workflow data for plotting by workflow id

    .. http:get:: /executions/data/<int: workflow_id>

    :param workflow_id: workflow id
    """

    workflow = db.session.get(WorkflowRun, workflow_id)
    if not workflow:
        return jsonify({})

    phase_data = {}
    main_phases = WorkflowPhase.query.filter_by(run_id=workflow_id, name='main') \
        .order_by(WorkflowPhase.repeat_index).all()

    for phase in main_phases:
        outputs = phase.outputs or {}
        phase_index = phase.repeat_index
        phase_data[phase_index] = {}

        # Normalize everything to a list of dicts
        if isinstance(outputs, dict):
            outputs = [outputs]
        elif isinstance(outputs, list):
            # flatten if it’s nested like [[{...}, {...}]]
            outputs = [
                item for sublist in outputs
                for item in (sublist if isinstance(sublist, list) else [sublist])
            ]

        # convert each output entry to plotting format
        for out in outputs:
            if not isinstance(out, dict):
                continue
            for k, v in out.items():
                if isinstance(v, (int, float)):
                    phase_data[phase_index].setdefault(k, []).append(
                        {"x": phase_index, "y": v}
                    )
                elif isinstance(v, list) and all(isinstance(i, (int, float)) for i in v):
                    phase_data[phase_index].setdefault(k, []).extend(
                        {"x": phase_index, "y": val} for val in v
                    )

    return jsonify(phase_data)


@data.delete("/executions/records/<int:workflow_id>")
@login_required
def delete_workflow_record(workflow_id: int):
    """
    .. :quickref: Workflow Data Database; delete a workflow execution record

    delete a workflow execution record by workflow id

    .. http:delete:: /executions/records/<int: workflow_id>

    :param workflow_id: workflow id
    :type workflow_id: int
    :status 200: return success message
    """
    run = WorkflowRun.query.get(workflow_id)
    db.session.delete(run)
    db.session.commit()
    return jsonify(success=True)


@data.route('/files/execution-data/<string:filename>')
@login_required
def download_results(filename:str):
    """
    .. :quickref: Workflow data; download a workflow data file (.CSV)

    .. http:get:: /files/execution-data/<string:filename>

    :param filename: workflow data filename
    :type filename: str

    # :status 302: load pseudo deck and then redirects to :http:get:`/ivoryos/executions/records`
    """

    filepath = os.path.join(current_app.config["DATA_FOLDER"], filename)
    return send_file(os.path.abspath(filepath), as_attachment=True)